import logging
import hashlib
import pathlib
from functools import wraps
from itertools import chain
import augpathlib as aug
from dateutil import parser
from augpathlib import PrimaryCache, EatCache, SqliteCache, SymlinkCache
from augpathlib import RepoPath, LocalPath
from sparcur import backends
from sparcur import exceptions as exc
from sparcur.utils import log, logd, GetTimeNow, register_type, unicode_truncate
from sparcur.utils import transitive_dirs, transitive_paths, is_list_or_tuple
from sparcur.utils import levenshteinDistance
from sparcur.utils import BlackfynnId, LocId, PennsieveId
from sparcur.config import auth


# http://fileformats.archiveteam.org/wiki/Main_Page
# https://www.nationalarchives.gov.uk/PRONOM/Format/proFormatSearch.aspx?status=new
# FIXME case senstivity
suffix_mimetypes = {
    ('.jpx',): 'image/jpx',
    ('.jp2',): 'image/jp2',

    # sequencing
    ('.fq',):           'application/fastq',  # wolfram lists chemical/seq-na-fastq which is overly semantic
    ('.fq', '.gz'):     'application/x-gz-compressed-fastq',
    ('.fastq',):        'application/fastq',  # XXX not official
    ('.fastq', '.gz'):  'application/x-gz-compressed-fastq',
    ('.fastq', '.bz2'): 'application/x-bzip-compressed-fastq',
    ('.bam',):          'application/x.vnd.hts.bam',

    ('.mat',):          'application/x-matlab-data',  # XXX ambiguous, depending on the matlab version
    ('.m',):            'application/x-matlab',

    ('.nii',):       'image/nii',  # XXX not official
    ('.nii', '.gz'): 'image/gznii',  # I'm not sure that I believe this # XXX not official

    # http://ced.co.uk/img/TrainDay.pdf page 7
    ('.smr',):       'application/x.vnd.cambridge-electronic-designced.spike2.32.data',
    ('.smrx',):      'application/x.vnd.cambridge-electronic-designced.spike2.64.data',

    ('.s2r',):       'application/x.vnd.cambridge-electronic-designced.spike2.resource',
    ('.s2rx',):      'application/x.vnd.cambridge-electronic-designced.spike2.resource+xml',

    ('.abf',):       'application/x.vnd.axon-instruments.abf',

    ('.nd2',):       'image/x.vnd.nikon.nd2',
    ('.czi',):       'image/x.vnd.zeiss.czi',
}

banned_basenames = (
    '.DS_Store',
    'Thumbs.db',
)


def cleanup(func):
    @wraps(func)
    def inner(self):
        meta = func(self)
        if meta is not None:
            if meta.id.startswith('EUf5'):
                if self.is_broken_symlink():
                    self.unlink()
                elif self.exists() and not self.is_symlink():
                    self.unlink()
                    self.touch()
                else:
                    pass

            else:
                return meta
            
    return inner


class BFPNCacheBase(PrimaryCache, EatCache):

    _backup_cache = SqliteCache
    _not_exists_cache = SymlinkCache
    _actually_crumple = False

    cypher = hashlib.sha256  # set the remote hash cypher on the cache class

    _suffix_mimetypes = suffix_mimetypes

    @classmethod
    def decode_value(cls, field, value):
        if field in ('created', 'updated'):
            # if you get unicode decode error here it is because
            # struct packing of timestamp vs isoformat are fighting
            # in xattrs pathmeta helper
            return value.decode()  # path meta handles decoding for us

    @property
    def anchor(self):
        return self.organization

    @property
    def identifier(self):
        ntid = self.id
        if ntid is None:
            raise exc.NoCachedMetadataError(self)

        return self._id_class(ntid, file_id=self.file_id)

    @property
    def _fs_safe_id(self):
        id = self.identifier
        return id.type[0] + '-' + id.uuid

    def crumple(self):
        """Avoid creating massive numbers of inodes by trashing paths during
           sync of old datasets."""
        if self.__class__._actually_crumple:
            super().crumple()
        else:
            if self.is_dir():
                self.rmdir()  # the dir has to be empty
            else:
                self.unlink()

    @property
    def _trashed_path(self):
        id = self.identifier
        suuid = id.uuid
        sid = id.type[0] + '-' + suuid  # must match _fs_safe_id
        try:
            pid = self.parent._fs_safe_id
        except exc.NoCachedMetadataError as e:
            msg = f'Projects cannnot be trashed when trying to trash {self}'
            raise TypeError(msg) from e

        fid = f'{self.file_id}-' if self.file_id else ''
        # we use sid as the sparsification folder here to ensure
        # a more uniform distribution of deletions across folders
        # therefore if you need to find files deleted from a specific folder
        # use find -name '*collection-id-part*'

        # nearly every file system we are likely to work with has 255 BYTE
        # limit on filename length, that is NOT CHAR length, so we have to
        # encode and then decode and truncate
        name = unicode_truncate(f'{sid}-{pid}-{fid}{self.name}', 255)

        return self.trash  / suuid[:2] / name

    @property
    def _trashed_path_short(self):
        return self.trash / 'short' / self.name

    @property
    def trash(self):
        return self.local_data_dir / 'trash'

    @property
    def organization(self):
        """ organization represents a permissioning boundary
            for blackfynn, so above this we would have to know
            in advance the id and have api keys for it and the
            containing folder would have some non-blackfynn notes
            also it seems likely that the same data could appear in
            multiple different orgs, so that would mean linking locally
        
            NOTE: unlike RemotePath, CachePath should use the local file
            structure to search for the local anchor.
        """

        # FIXME in reality files can have more than one org ...
        if self.meta is not None:
            identifier = self.identifier
            if identifier is None:
                parent = self.parent
                if parent == self:  # we have hit a root
                    return None

                organization = parent.organization

                if organization is not None:
                    # TODO repair
                    pass
                else:
                    raise exc.NotInProjectError

            elif identifier.type == 'organization':
                return self

        if self.parent and self.parent != self:
            # we have a case of missing metadata here as well
            return self.parent.organization

    def is_organization(self):  # FIXME ?
        return self.id and self.identifier.type == 'organization'

    def is_dataset(self):
        return self.id and self.identifier.type == 'dataset'

    @property
    def dataset(self):
        if self.is_dataset():
            return self

        elif self.parent and self.parent != self:  # Path('.') issue
            log.log(9, self.parent)
            return self.parent.dataset

    @property
    def dataset_id(self):
        dataset = self.dataset
        if dataset:
            return dataset.id

    @property
    def file_id(self):
        return self.meta.file_id

    @property
    def cache_key(self):
        # FIXME file system safe
        id = self.identifier
        uuid = id.uuid
        return f'{uuid[:2]}/{uuid}-{self.file_id}'

    def _dataset_metadata(self, force_cache=False):
        """ get metadata about a dataset from the remote metadata store """
        # TODO figure out the right default for force_cache
        dataset = self.dataset
        if dataset == self:
            if not hasattr(self, '_c__dataset_metadata'):
                pdd = backends.PennsieveDatasetData(self)
                try:
                    blob = pdd.fromCache()
                except FileNotFoundError as e:
                    # FIXME TODO also check cached rmeta dates during pull
                    if force_cache:
                        raise e
                    else:
                        log.warning(e)
                        blob = pdd()

                self._c__dataset_metadata = blob

            return self._c__dataset_metadata

        else:
            return dataset._dataset_metadata()

    def _package_count(self):
        if self.is_dataset():
            return sum(self._dataset_metadata()['package_counts'].values())
        else:
            raise NotImplementedError('unused at the moment')

    def _sparse_materialize(self, *args, sparse_limit=None):
        """ use data from the remote mark or clear datasets as sparse """
        if sparse_limit is None:
            sparse_limit = auth.get('sparse-limit')  # yay for yaml having int type

        if self.is_dataset():
            package_count = self._package_count()
            sparse_remote = (
                False
                if sparse_limit is None or sparse_limit < 0 else
                package_count >= sparse_limit)
            sparse_cache = self.is_sparse()
            if sparse_remote:
                if not sparse_cache:
                    self._mark_sparse()
            elif sparse_cache:  # strange case where number of packages decreases
                self._clear_sparse()

        else:
            msg = 'at the moment only datasets can be marked as sparse'
            raise NotImplementedError(msg)

    @property
    def data(self):
        """ get the 'cached' data which isn't really cached at the moment
            once we implement an index for local files then we can hit that
            first from here """
        # we don't keep two copies of the local data
        # unless we are doing a git-like thing
        if self.is_dir():
            raise TypeError('can\'t retrieve data for a directory')

        meta = self.meta
        if meta.file_id is None:
            raise NotImplementedError('can\'t fetch data without a file id')

        #cands = list(self.local_object_cache_dir.glob(self.cache_key))
        # FIXME this does not play well with old_id ...
        # can probably get away with just globing for the old_id in
        # most cases
        # TODO where to store the chain of prior versions? i.e. do
        # we just keep the xattrs in the object cache? how about file moves?
        # sigh git ...
        rgen = None
        if self.local_object_cache_path.exists():
            locsize = self.local_object_cache_path.size
            if locsize != meta.size:
                msg = (f'Partial download detected {locsize} != {meta.size} at'
                       f'\n{self.local_object_cache_path}')
                log.info(msg)
                size = self.local_object_cache_path.size
                kwargs = {}
                if size > 0:
                    if (self.local == self.local_object_cache_path
                        and size > 4096):  # FIXME hardcoded chunksize
                        # XXX there is a fantastic edge case where if
                        # you try to read and write from the same file
                        # only the first chunk will be written and if
                        # you are retrieving from remote then the offset
                        # would be greater than the chunksize so there
                        # will be a gap, so we set chunksize here and
                        # issue a critical log
                        msg = ('You probably did not mean to do this. '
                               f'Refetching {size - 4096} bytes.')
                        log.critical(msg)
                        kwargs['ranges'] = ((4096,),)
                    else:
                        kwargs['ranges'] = ((size,),)

                if not hasattr(self._remote_class, '_api'):
                    # see note below
                    self._remote_class.anchorToCache(self.anchor)

                # make sure we can write to locp since cached objects
                # are set readonly on completion
                self.local_object_cache_path.chmod(0o0644)

                rgen = self._remote_class.get_file_by_id(
                    meta.id, meta.file_id, **kwargs)
                gen = chain(
                    (next(rgen),),
                    self.local_object_cache_path.data)
            else:
                gen = chain(
                    (f'from local cache {self.local_object_cache_path}',),
                    self.local_object_cache_path.data)
        else:
            if not hasattr(self._remote_class, '_api'):
                # NOTE we do not want to dereference self.remote
                # in this situation because we just want the file
                # not the FS metadata, so we have to ensure that _api
                # is bound
                self._remote_class.anchorToCache(self.anchor)

            gen = self._remote_class.get_file_by_id(meta.id, meta.file_id)

        try:
            self.data_headers = next(gen)
        except exc.NoRemoteFileWithThatIdError as e:
            log.error(f'{self} {e}')
            raise exc.CacheNotFoundError(f'{self}') from e  # have to raise so that we don't overwrite the file

        log.log(9, self.data_headers)
        if self.local_object_cache_path.exists():
            yield from gen
            if rgen is None:
                return

            yield from self.local_object_cache_path._data_setter(
                rgen, append=True)

        else:
            # FIXME we MUST write the metadata first so that we know the expected size
            # so that in the event that the generator is only partially run out we know
            # that we can pick up where we left off with the fetch, this also explains
            # why all the cases where the cached data size did not match were missing
            # xattrs entirely
            _lcp = self.local_object_cache_path.parent
            if not _lcp.exists():
                # FIXME sigh, no obvious way around having to check
                # every time other than creating all the cache
                # subfolders in advance
                # XXX SOMEHOW between the call in the if statement above
                # this call right here it is possible for something running
                # in async in pypy3 to manage to have another threadthing
                # create the directory ... this is beyond annoying
                _lcp.mkdir(exist_ok=True)

            self.local_object_cache_path.touch()
            self.local_object_cache_path.cache_init(meta)

            yield from self.local_object_cache_path._data_setter(gen)

        # ensure we don't accidentally modify cached files if they are symlinked
        self.local_object_cache_path.chmod(0o0444)  # unlink works on 444

        ls = self.local_object_cache_path.size
        if ls != meta.size:
            self.local_object_cache_path.unlink()
            msg = f'{ls} != {meta.size} for {self}'
            raise ValueError(msg)  # FIXME TODO

    def _meta_is_root(self, meta):
        return meta.id.startswith('N:organization:')

    def _bootstrap_recursive(self, only=tuple(), skip=tuple(), sparse=False):
        """ only on the first call to this function should sparse be a tuple """
        # bootstrap the rest if we told it to
        if self.id.startswith('N:organization:'):  # FIXME :/
            yield from self.remote.children_pull(self.children,
                                                 only=only,
                                                 skip=skip,
                                                 sparse=sparse,)

        else:
            if isinstance(sparse, list) and self.id.startswith('N:dataset:'):
                sparse = self.id in sparse

            yield from super()._bootstrap_recursive(sparse=sparse)

    _sparse_stems = (
        'manifest', 'dataset_description', 'submission', 'performances', 'subjects', 'samples'
    )

    _sparse_exts = (
        'xlsx', 'csv', 'tsv', 'json'
    )

    def _sparse_include(self):
        sl = self.stem.lower()
        return bool([an for an in self._sparse_stems if an in sl])

    def populateJsonMetadata(self, blob):
        """ populate a json blob with file system metadata"""
        # TODO schema for this
        if blob is None:
            blob = {}

        blob.update(self._jsonMetadata())

    def _jsonMetadata(self, do_expensive_operations=False):
        """ path (cache) level json metadata """
        # FIXME this is going to be very slow due to the local.cache call overhead
        # it will be much better implement this from Path directly using xattrs()
        d = self.dataset
        l = self.local
        drp = l.relative_to(d.local)
        meta = self.meta  # TODO see what we need from this
        blob = {
            'type': 'path',
            'dataset_id': d.identifier.curie,
            'dataset_relative_path': drp,
            'uri_api': self.uri_api,  # -> @id in the @context
            'uri_human': self.uri_human,
            }
        if hasattr(self, '_remote_class'):  # FIXME hack to detect CacheL cases
            blob['remote_id'] = (
                # XXX a horrible hack, but it kinda works
                f'{self.identifier.curie}/files/{meta.file_id}'
                if self.is_file() or self.is_broken_symlink() else
                self.identifier.curie)

        # FIXME N:package:asdf is nasty for jsonld but ...
        # yes the bf_id will have to be parsed to know what
        # endpoint to send it to ... yay for scala thinking amirite !? >_<
        # but at least for this particular structure it

        # NOTE packages are collections of size 1 > more
        # and and when they happend to have only a single
        # member they are conflated with the single file
        # they contain

        if meta.checksum is not None:
            blob['checksums'] = [{'type': 'checksum',
                                  # FIXME cypher should ALWAYS travel with checksum
                                  # not be implicit and based on the implementation
                                  'cypher': self.cypher.__name__.replace('openssl_', ''),
                                  'hex': meta.checksum.hex(),}]

        mimetype = l.mimetype
        if mimetype:
            blob['mimetype'] = mimetype

        if do_expensive_operations:
            blob['magic_mimetype'] = asdf  # FIXME NameError

        return blob


class BlackfynnCache(BFPNCacheBase, PrimaryCache, EatCache):

    xattr_prefix = 'bf'

    uri_human = backends.BlackfynnRemote.uri_human
    uri_api = backends.BlackfynnRemote.uri_api

    _id_class = BlackfynnId


BlackfynnCache._bind_flavours()


class PennsieveCache(BFPNCacheBase, PrimaryCache, EatCache):

    #xattr_prefix = 'pn'  # FIXME TODO we will likely need an xattr conversion fix
    xattr_prefix = 'bf'

    uri_human = backends.PennsieveRemote.uri_human
    uri_api = backends.PennsieveRemote.uri_api

    _id_class = PennsieveId


PennsieveCache._bind_flavours()


class Path(aug.XopenPath, aug.RepoPath, aug.LocalPath):  # NOTE this is a hack to keep everything consistent
    """ An augmented path for all the various local
        needs of the curation process. """

    _cache_class = BlackfynnCache

    _dataset_cache = {}

    _suffix_mimetypes = suffix_mimetypes

    @classmethod
    def fromJson(cls, blob):
        # FIXME the problem here is that we can't distingish between cases where
        # the original path was absolute, vs cases where we are embedding this data
        # the the path was relative to start with
        return blob

        #return Path(blob['dataset_relative_path'])  # doesn't work, want full path
        if (hasattr(cls, '_cache_class') and
            hasattr(cls._cache_class, '_anchor') and
            cls._cache_class._anchor is not None):
            a = cls._cache_class._anchor
            #l = a.local
            did = 'N:' + blob['dataset_id']
            if did not in cls._dataset_cache:
                cls._dataset_cache.update({c.id:c.local for c in a.children})

            dataset_local = cls._dataset_cache[did]
            path_relative_string = blob['dataset_relative_path']
            return dataset_local / path_relative_string
        else:
            return blob  # FIXME TODO

    @property
    def project_relative_path(self):
        return self.relative_path_from(self._cache_class._anchor.local)

    @property
    def dataset_relative_path(self):
        # FIXME broken for local validation
        # update: I'm pretty sure this works for total local validation
        # because we use CacheL, so the fix here is all we need to cover
        # the file changed case
        try:
            if self.cache is None and self.parent.cache is not None:
                # file has been modified case
                return self.relative_path_from(self.parent.cache.dataset.local)
            else:
                return self.relative_path_from(self.cache.dataset.local)
        except AttributeError as e:
            raise exc.NoCachedMetadataError(self) from e

    @property
    def cache_id(self):
        """ fast way to get cache.id terrifyingly this has roughly
        an order of magnitude less overhead than self.cache.id """

        try:
            return (self.getxattr('bf.id').decode()
                    if self.is_dir() or self.is_file() else
                    self.readlink().parts[1])
        except FileNotFoundError as e:
            # in the event we try to blindly get cached metadata for a
            # path in a manifest that does not exist
            raise e
        except OSError as e:
            raise exc.NoCachedMetadataError(self) from e

    @property
    def cache_file_id(self):
        if self.is_dir():
            return

        try:
            return (self.getxattr('bf.file_id').decode()
                    if self.is_file() else
                    self.cache_meta.file_id)
        except FileNotFoundError as e:
            # in the event we try to blindly get cached metadata for a
            # path in a manifest that does not exist
            raise e
        except OSError as e:
            raise exc.NoCachedMetadataError(self) from e

    @property
    def cache_identifier(self):
        return self._cache_class._id_class(
            self.cache_id, file_id=self.cache_file_id)

    @property
    def rchildren_dirs(self):
        # FIXME windows support if find not found
        if self.is_dir():
            yield from transitive_dirs(self)

    def populateJsonMetadata(self, blob):
        """ populate a json blob with file system metadata"""
        # FIXME TODO implement this without resolving .cache
        if self.cache is not None:
            return self.cache.populateJsonMetadata(blob)
        else:
            return self._jsonMetadata()

    xattr_prefix = BlackfynnCache.xattr_prefix
    _xattr_meta = aug.EatCache.meta
    _symlink_meta = aug.SymlinkCache.meta

    @property
    def cache_meta(self):
        """ Access cached metadata without constructing the cache instance. """
        if self.is_broken_symlink():
            # NOTE check ibs first otherwise symlink meta could
            # raise a NoCachedMetadataError, it may raise other
            # errors if something goes wrong when reading the
            # symlink, but not that particular error
            return self._symlink_meta
        else:
            return self._xattr_meta

    def manifest_record(self, manifest_parent_path):
        filename = self.relative_path_from(manifest_parent_path)
        description = None
        filetype = self.mimetype  # TODO failover
        additional_type = None  # TODO
        if self.is_file():
            meta = self.meta
            # use updated because most file systems don't have a
            # meaningful created time aka birthdate
            timestamp = meta.updated
            checksum = meta.checksum
        else:
            meta = self.cache.meta
            timestamp = meta.created  # XXX created on remote platform time
            checksum = None  # XXX avoid promulgating unchecked values

        return (
            filename,
            timestamp,
            description,
            filetype,
            additional_type,
            checksum,
        )

    def generate_manifest(self, include_directories=False):
        """ generate a tabular manifest of all contents of a directory
            serialization is handled by the caller if it is required """

        if not self.is_dir():
            log.error('Can only generate manifests for directories!')
            raise NotADirectoryError(self)

        if include_directories:
            return [c.manifest_record(self) for c in self.rchildren]
        else:
            return [c.manifest_record(self) for c in self.rchildren
                    if not c.is_dir()]

    def _cache_jsonMetadata(self, do_expensive_operations=False):
        blob, project_path, dsid = self._jm_common(do_expensive_operations=do_expensive_operations)
        project_meta = project_path.cache_meta
        meta = self.cache_meta  # FIXME this is very risky
        # and is the reason why I made it impossible to contsturct
        # cache classes when no cache was present, namely that if
        # there is no cache then PathMeta will still return the
        # correct structure, but it will be empty, which is bad
        if meta is None:
            log.critical(f'something is very wrong with path: {self}')
            raise exc.NoCachedMetadataError(self)
        elif meta.id is None:
            raise exc.NoCachedMetadataError(self)

        _, bf_id = meta.id.split(':', 1)  # FIXME SODA use case
        idc = self._cache_class._id_class
        remote_id = (idc(bf_id, file_id=meta.file_id)
                     if self.is_file() or self.is_broken_symlink() else
                     idc(bf_id))
        #identifier = self._cache_class._id_class('N:' + remote_id)
        uri_api = remote_id.uri_api
        uri_human = remote_id.uri_human(
            organization=project_path.cache_identifier,
            dataset=dsid)
        blob['remote_id'] = remote_id
        #blob['timestamp_created'] = meta.created  # leaving this out
        blob['timestamp_updated'] = meta.updated  # needed to simplify transitive update
        blob['uri_api'] = uri_api
        blob['uri_human'] = uri_human

        if meta.size is not None:
            blob['size_bytes'] = meta.size

        if (self.is_file() or self.is_broken_symlink()) and meta.checksum is not None:
            # FIXME known checksum failures !!!!!!!
            blob['checksums'] = [{'type': 'checksum',
                                  # FIXME cypher should ALWAYS travel with checksum
                                  # not be implicit and based on the implementation
                                  'cypher': self._cache_class.cypher.__name__.replace('openssl_', ''),
                                  'hex': meta.checksum.hex(),}]

        return blob

    def _jm_common(self, do_expensive_operations=False):
        # FIXME WARNING resolution only works if we were relative to
        # the current working directory
        if self.is_broken_symlink():
            self = self.absolute()
        else:
            self = self.resolve()  # safety since we have to go hunting paths

        project_path = self.find_cache_root()

        if project_path is None:
            # FIXME TODO I think we used dataset_description as a hint?
            project_path = self.__class__('/')  # FIXME FIXME
            log.critical(f'No dataset path found for {self}!')
            #raise NotImplementedError('current implementation cant anchor with current info')

        dataset_path = [p for p in chain((self,), self.parents) if p.parent == project_path][0]
        drp = self.relative_path_from(dataset_path)  # FIXME ...

        dsid = dataset_path.cache_identifier

        blob = {
            'type': 'path',
            'dataset_id': dsid.curie,
            'dataset_relative_path': drp,
            'basename': self.name,  # for sanity's sake
        }

        mimetype = self.mimetype
        if mimetype:
            blob['mimetype'] = mimetype

        if do_expensive_operations:
            blob['magic_mimetype'] = self._magic_mimetype

        if not (self.is_broken_symlink() or self.exists()):
            # TODO search for closest match
            cands = self._closest_existing_matches()
            msg = f'Path does not exist!\n{self}'
            if cands:
                _fcands = [(r, n) for r, n in cands if r < 10]
                fcands = _fcands if _fcands else cands
                msg += f'\n{0: <4} {self.name}\n'
                msg += '\n'.join([f'{r: <4} {n}' for r, n in fcands])
            # do not log the error here, we won't have
            # enough context to know where we got a bad
            # path, but the caller should, maybe a case for
            # inversion of control here
            blob['errors'] = [{'message': msg,
                               'blame': 'submissions',  # FIXME maybe should be pipeline?
                               'pipeline_stage': 'sparcur.path._jm_common',
                               'candidates': cands,}]

        return blob, project_path, dsid

    def _closest_existing_matches(self):
        if self.parent.exists():
            # we probably don't need this, the numbers
            # should usually be small enough
            #childs = [c for c in self.parent.children
                      #if c.is_file() or c.is_broken_symlink()]
            name = self.name
            cands = sorted(
                [(levenshteinDistance(c.name, name), c.name)
                 for c in self.parent.children])
            return cands

    def _jsonMetadata(self, do_expensive_operations=False):
        blob, project_path, dsid = self._jm_common(do_expensive_operations=do_expensive_operations)
        return blob

    def _transitive_metadata(self):
        def hrm(p):
            try:
                cjm = p._cache_jsonMetadata()
            except exc.NoCachedMetadataError:
                # FIXME TODO figure out when to log this
                # probably only log when there is a dataset id
                return
            # this is not always needed so it is not included by default
            # but we do need it here for when we generate the rdf export
            p = cjm['dataset_relative_path'].parent
            if p.name != '':  # we are at the root of the relative path aka '.'
                cjm['parent_drp'] = p
            return cjm

        def sort_dirs_first(d):
            # NOTE THAT mimetype IS NOT A REQUIRED FIELD
            return ((not ('mimetype' in d and
                          d['mimetype'] == 'inode/directory')),
                    'mimetype' in d)

        exclude_patterns = '*.~lock*#', '.DS_Store'  # TODO make this configurable
        _tm = hrm(self)
        _tm['dataset_relative_path'] = ''  # don't want . since it is ambiguous?
        this_metadata = [_tm]  # close the transitive metadata under parent_id
        # FIXME perf boost but not portable
        rchildren = transitive_paths(self, exclude_patterns=exclude_patterns)
        _rcm = [_ for _ in [hrm(c) for c in rchildren] if _ is not None]
        rc_metadata = sorted(_rcm, key=sort_dirs_first)
        metadata = this_metadata + rc_metadata
        index = {b['dataset_relative_path']:b for b in metadata}
        pid = this_metadata[0]['remote_id']
        for m in metadata:
            p = m.pop('parent_drp', None)
            if p in index:
                m['parent_id'] = index[p]['remote_id']
            else:
                m['parent_id'] = pid

        return metadata

    @classmethod
    def _file_type_status_lookup(cls):
        import json  # FIXME
        if not hasattr(cls, '_sigh_ftslu'):
            resources = auth.get_path('resources')  # FIXME breaks outside git
            with open(resources / 'mimetypes.json', 'rt') as f:
                classification = json.load(f)

            mimetypes = {mimetype:status for status,objs in
                         classification.items() for obj in objs
                         for mimetype in (obj['mimetype']
                                          if is_list_or_tuple(obj['mimetype']) else
                                          (obj['mimetype'],))}
            suffixes = {obj['suffix']:status for status,objs in
                        classification.items() for obj in objs}
            cls._mimetypes_lu, cls._suffixes_lu = mimetypes, suffixes
            cls._sigh_ftslu = True


        return cls._mimetypes_lu, cls._suffixes_lu

    _banned_basenames = banned_basenames
    _unknown_suffixes = set()
    _unclassified_mimes = set()
    @classmethod
    def validate_path_json_metadata(cls, path_meta_blob):
        from sparcur.core import HasErrors  # FIXME
        he = HasErrors(pipeline_stage=cls.__name__ + '.validate_path_json_metadata')
        mimetypes, suffixes = cls._file_type_status_lookup()  # SIGH this overhead is 2 function calls and a branch
        for i, path_meta in enumerate(path_meta_blob['data']):
            if path_meta['basename'] in cls._banned_basenames:
                msg = f'illegal file detect {path_meta["basename"]}'
                dsrp = path_meta['dataset_relative_path']
                if he.addError(msg, path=dsrp, json_path=('data', i)):
                    logd.error(msg)
                status = 'banned'
                path_meta['status'] = status
                continue

            if 'magic_mimetype' in path_meta and 'mimetype' in path_meta:
                # FIXME NOT clear whether magic_mimetype should be used by itself
                # usually magic and file extension together work, magic by itself
                # can give some completely bonkers results
                source = 'magic_mimetype'
                mimetype = path_meta['magic_mimetype']
                muggle_mimetype = path_meta['mimetype']
                if mimetype != muggle_mimetype:
                    msg = f'mime types do not match {mimetype} != {muggle_mimetype}'
                    dsrp = path_meta['dataset_relative_path']
                    if he.addError(msg, path=dsrp, json_path=('data', i)):
                        log.error(msg)
            elif 'magic_mimetype' in path_meta:
                source = 'magic_mimetype'
                mimetype = path_meta['magic_mimetype']
            elif 'mimetype' in path_meta:
                source = 'mimetype'
                mimetype = path_meta['mimetype']
            else:
                mimetype = None

            if mimetype is not None:
                try:
                    status = mimetypes[mimetype]
                    if status == 'banned':
                        msg = f'banned mimetype detected {mimetype}'
                        dsrp = path_meta['dataset_relative_path']
                        if he.addError(msg, path=dsrp,
                                       json_path=('data', i, source)):
                            logd.error(msg)
                except KeyError as e:
                    status = 'known'
                    if mimetype not in cls._unclassified_mimes:
                        cls._unclassified_mimes.add(mimetype)
                        log.info(f'unclassified mimetype {mimetype}')
            else:
                status = 'unknown'
                dsrp = path_meta['dataset_relative_path']
                if isinstance(dsrp, str):
                    if not dsrp:
                        msg = f'FIXME top level folder needs a mimetype!'
                    else:
                        msg = f'unknown mimetype {path_meta["basename"]}'
                else:
                    msg = f'unknown mimetype {"".join(dsrp.suffixes)}'
                    cls._unknown_suffixes.add(tuple(dsrp.suffixes))
                if he.addError(msg, path=dsrp, json_path=('data', i)):
                    logd.warning(msg)

            path_meta['status'] = status

        if he._errors_set:
            he.embedErrors(path_meta_blob)

    def pull(self, *args,
             paths=None,
             time_now=None,
             debug=False,
             n_jobs=12,
             cache_anchor=None,
             log_name=None,
             log_level='INFO',
             # pass in Parallel in at run time if needed
             Parallel=None,
             delayed=None,
             _in_parallel=False,
             exclude_uploaded=False,):
        # TODO usage errors

        if time_now is None:
            time_now = GetTimeNow()
            log.debug('No time provided to pull so using '
                      f'{time_now.START_TIMESTAMP}')

        if _in_parallel:
            _log = logging.getLogger(log_name)
            _log.setLevel(log_level)
            rc = self._remote_class
            if not hasattr(rc, '_cache_anchor'):
                rc.anchorTo(cache_anchor)

        else:
            _log = log

        cache = self.cache

        if cache.is_organization():
            if debug or Parallel is None or n_jobs <= 1:
                for child in self.children:
                    if paths is None or child in paths:
                        child.pull()
            else:
                Parallel(n_jobs=n_jobs)(
                    delayed(child.pull)(_in_parallel=True,
                                        time_now=time_now,
                                        cache_anchor=cache.anchor,
                                        log_name=_log.name,
                                        log_level=log_level,
                                        exclude_uploaded=exclude_uploaded,)
                    for child in self.children
                    if paths is None or child in paths)

        elif cache.is_dataset():
            self._pull_dataset(time_now, exclude_uploaded)  # XXX actual pull happens in here

        else:
            raise NotImplementedError(self)

    def _pull_dataset(self, time_now, exclude_uploaded):
        cache = self.cache
        try:
            _old_eu = self._remote_class._exclude_uploaded
            cache._remote_class._exclude_uploaded = exclude_uploaded
            return self._pull_dataset_internal(time_now)
        finally:
            cache._remote_class._exclude_uploaded = _old_eu

    def _pull_dataset_internal(self, time_now):
        cache = self.cache
        if cache.is_organization():
            raise TypeError('can\' use this method on organizations')
        elif not cache.is_dataset():
            return self.dataset._pull_dataset_internal(time_now)

        children = list(self.children)
        if not children:
            return list(self.cache.rchildren)  # XXX actual pull happens here

        # instantiate a temporary staging area for pull
        ldd = cache.local_data_dir
        contain_upstream = ldd / 'temp-upstream'
        contain_upstream.mkdir(exist_ok=True)

        # create the new directory with the xattrs needed to pull
        working = self
        upstream = contain_upstream / working.name
        try:
            upstream.mkdir()
            # if upstream fails that means that somehow a previous
            # cleanup did not succeed (for some reason), this time it
            # should succeed and next time everything should work
        except FileExistsError as e:
            log.exception(e)
            error_suf = f'{time_now.START_TIMESTAMP_LOCAL_SAFE}-ERROR'
            upstream.rename(upstream.parent / (upstream.name + error_suf))
            upstream.mkdir()
        except BaseException as e:
            error_suf = f'{time_now.START_TIMESTAMP_LOCAL_SAFE}-ERROR'
            upstream.rename(upstream.parent / (upstream.name + error_suf))
            # TODO do we need to log here or can we log up the stack?
            raise e

        # if mkdir succeeds then any failure that happens between
        # then and the point where we swap needs to move the folder
        # to a unique path marked as an error so that future calls
        # to this function are not stopped by an existing directory
        # that failed to be cleaned up
        upstream.setxattrs(working.xattrs())  # handy way to copy xattrs
        # pull and prep for comparison of relative paths
        upstream_c_children = list(upstream.cache.rchildren)  # XXX actual pull happens here
        upstream_children = [c.local.relative_to(upstream)
                                for c in upstream_c_children]
        working_children = [c.relative_to(working)
                            for c in working.rchildren]

        # FIXME TODO comparison/diff and support for non-curation workflows

        # replace the working tree with the upstream
        # FIXME this has nasty and confusing consequences if
        # working == aug.AugmentedPath.cwd()
        upstream.swap_carefree(working)

        upstream_now_old = upstream

        # neither of these cleanup approaches is remotely desireable
        suf = f'-{time_now.START_TIMESTAMP_LOCAL_SAFE}'
        old = upstream_now_old.parent / (upstream_now_old.name + suf)
        upstream_now_old.rename(old)  # FIXME
        # TODO cleanup old either by removing things older than X
        # or by symlinking unmodified files to their package cache
        old._symlink_packages_transitive()

    def _eq_data(self, other):
        for schunk, ochunk in zip(self.data, other.data):
            if schunk != ochunk:
                return False

        return True

    def _symlink_packages_transitive(self, remove_circular=True):
        # the ext4 defaults on are at least an order of magnitude too low to
        # retain all the broken symlinks for thousands of snapshots, we also
        # don't particularly need to keep them around

        # find all files
        # if real file
        # determine if they differ from the cache XXX this is going to be slow right now
        # if not changed remove the file in the tree and symlink it to the package
        # if changed leave the modified file
        # if broken symlink unlink to keep inodes under control
        for c in self.rchildren:
            if c.cache is not None:  # FIXME need a way to recover the original package?
                locp = c.cache.local_object_cache_path
                if (c.is_file() and locp.exists() and
                    c.size == locp.size and
                    c._eq_data(locp)):
                    c.unlink()
                    relsym = c.relative_path_to(locp)
                    c.symlink_to(relsym)
                if remove_circular and c.is_broken_symlink():
                    c.unlink()

    def updated_cache_transitive(self):
        """ fast get the date for the most recently updated cached path """
        if self.cache.is_organization():
            gen = (rc for c in self.children for rc in c.rchildren)
        elif self.cache.is_dataset():
            gen = self.rchildren
        else:
            gen = chain((self,), self.rchildren)

        def updated_hierarchy(local):
            """ try to get the cache updated time, if the cache doesn't exist,
                e.g. because the file has been overwritten and there are no
                xattrs, then use the local updated time """
            try:
                updated = local.getxattr('bf.updated').decode()
                return aug.PathMeta(updated=updated)
            except exc.NoStreamError:
                return local.meta_no_checksum


        simple_meta = [updated_hierarchy(c)
                       if not c.is_broken_symlink() else
                       aug.PathMeta.from_symlink(c)
                       for c in gen]
        if simple_meta:
            return max(m.updated for m in simple_meta)
        else:
            # in the even that the current folder is empty
            updated = self.getxattr('bf.updated').decode()
            return aug.PathMeta(updated=updated).updated

    def _upload_raw(self):
        remote, old_remote = (
            self
            ._cache_class
            ._remote_class
            ._stream_from_local_raw(self))

    def upload(self, replace=True, local_backup=False):
        # FIXME we really need a staging area ...
        remote, old_remote = (
            self
            ._cache_class
            ._remote_class
            ._stream_from_local(self,
                                replace=replace,
                                local_backup=local_backup))

        if True:  # local_backup = True:  # force True until we can get remote checksums
            # FIXME there must be a better way to do this ...
            object_path = self._cache_class._anchor.local_objects_dir / remote.cache_key
            # TODO possibly use _data_setter to calculate the hash at the same time?
            if not object_path.parent.exists():
                # XXX FIXME multi-level cache keys are a ticking timebomb
                object_path.parent.mkdir(parents=True)

            object_path.copy_from(self)
            object_path.cache_init(remote.meta)

        if self.cache is None:
            # FIXME didn't we already figure out the right way to do this?
            cache = self.cache_init(remote.meta)
            self._cache = cache
            cache._remote = remote
            remote._cache = cache
        else:
            # remote has no cache and self already has a cache in this case
            # so we can't just assign directly so we have to say which
            # cache we want to update, and make sure that names match
            # the cache should already have been updated EXCEPT for the
            # fact that the remote fileid is not known, but we are inside
            # the code that actually makes that update ...
            # and the absense of a staging area / history makes the cache.meta
            # the only place we can do this besides the object store, especially
            # if we lack the ability to retrieve checksums for single files (sigh)

            # we already have the latest data (ignoring concurency)
            remote.update_cache(cache=self.cache, fetch=False) # FIXME fetch=False => different diff rule

        return remote


Path._bind_flavours()
register_type(Path, 'path')


class RemoteL:

    _remote_type = 'local'
    _cache_anchor = None

    def __new__(cls, id, *args, **kwargs):
        return cls._local_class(id.split(':', 1)[1])

    @classmethod
    def anchorToCache(cls, cache, *args, **kwargs):
        cls._cache_anchor = cache


class CacheL(aug.caches.ReflectiveCache, EatCache):

    _id_class = LocId

    _remote_class = RemoteL
    _asserted_anchor = None
    _dataset_dirs = []

    _jsonMetadata = BFPNCacheBase._jsonMetadata
    populateJsonMetadata = BFPNCacheBase.populateJsonMetadata
    cypher = BFPNCacheBase.cypher

    def _type(self):
        if self.is_file():
            return 'package'
        elif self.is_dir():
            if 'organization' in self.id:
                # have to do this to prevent recursion
                return 'organization'
            elif self == self._asserted_anchor:
                return 'organization'
            elif self in self._dataset_dirs or self.parent.is_organization():
                return 'dataset'
            else:
                return 'collection'
        elif not self.exists():
            log.warning(f'File should exist? {self!r}!')
            raise FileNotFoundError(self)
        else:
            msg = f"unsupported inode type at path\n'{self.as_posix()}'"
            raise TypeError(msg)

    @property
    def identifier(self):
        return self._id_class(self.id, self._type())

    @property
    def uri_api(self):
        return self.identifier.uri_api

    @property
    def uri_human(self):
        return self.identifier.uri_human(
            dataset=self.dataset,
            organization=self.anchor)

    @property
    def anchor(self):
        if self._asserted_anchor is not None:
            return self._asserted_anchor
        elif self.identifier.type == 'organization':
            return self
        elif self.parent != self:
            return self.parent.anchor
        else:
            raise exc.NotInProjectError(f'{self} is not in a project!')

    dataset = BlackfynnCache.dataset

    dataset_id = BlackfynnCache.dataset_id

    def is_organization(self):
        return self._type() == 'organization'

    def is_dataset(self):
        return self._type() == 'dataset'


CacheL._bind_flavours()


class PathL(Path):

    # FIXME this seems like a good idea until you realize some of the
    # other classes that we built inherit from Path, because it
    # simplifies the implementation, but we actually want the type
    # specialization to depend on the type that we passed to it, which
    # requires modifying new to match the constructed class, sigh XXX
    # the solution is to overwrite the cache class on instances that
    # are subclasses of Path during construction, I think that is the
    # only way to be sure, clunky and aweful, but unavoidable

    _cache_class = CacheL
    # setting remote class is required to avoid issue with
    # PennsieveRemote -> Cache/Path instead of to CacheL/PathL

    def find_cache_root(self):
        # straight forward when the anchor is asserted
        return self.cache.anchor

    @property
    def cache_identifier(self):
        return self.cache.identifier


PathL._bind_flavours()
CacheL._local_class = PathL
RemoteL._local_class = PathL


class StashCache(BlackfynnCache):
    def remote(self):
        return None

    def _meta_setter(self, value):
        if not self.meta:
            super()._meta_setter(value)
        else:
            raise TypeError('you dont want to set a stashed cache')


class StashPath(Path):
    _cache_class = StashCache
    @property
    def remote(self):
        return None


# assign defaults

def bind_defaults(Remote, Cache):
    Cache._local_class = Path
    Remote._new(Path, Cache)
    Remote.cache_key = Cache.cache_key
    Remote._sparse_stems = Cache._sparse_stems
    Remote._sparse_exts = Cache._sparse_exts
    Remote._sparse_include = Cache._sparse_include


bind_defaults(backends.BlackfynnRemote, BlackfynnCache)
# XXX supersedes the binding on BlackfynnRemote
bind_defaults(backends.PennsieveRemote, PennsieveCache)



import ast
import copy
import json
import types
import inspect
from functools import wraps
import jsonschema

# FIXME these imports should not be here types rules should be set in another way
from pathlib import PurePath
import idlib
import rdflib
import requests
import ontquery as oq
from pysercomb.pyr.types import ProtcurExpression
from sparcur import exceptions as exc
from sparcur.utils import logd, BlackfynnId, PennsieveId
from sparcur.core import JEncode, JApplyRecursive, JPointer, OntCuries
from sparcur.core import get_nested_by_key
from pyontutils.utils import asStr
from pyontutils.namespaces import TEMP, TEMPRAW, sparc, unit, PREFIXES as uPREFIXES, ilxtr


CONTEXT_ROOT = object()  # used to mark where nested context_value statements should be deposited
prefix_endswith = ['/', '#', '_', '-', ':', '=',
                   '/fma', '/sao', 'PTHR', '/chebi#2', '/chebi#3',]

base_context = {
    '@version': 1.1,
    'id': '@id',
    'dataset': {'@id': 'https://api.pennsieve.io/datasets/N:dataset:', '@prefix': True},  # FIXME -> pennsieve
    'package': {'@id': 'https://api.pennsieve.io/packages/N:package:', '@prefix': True},  # FIXME -> pennsieve
    # @base and _bfc are added at runtime and are dataset:{dataset-id-suffix}
    'meta': '@graph',
    #'meta': 'TEMP:hasSubGraph',  #'_bfc:#meta-graph',  # FIXME I think the fragment is the right thing to do here ...
    # the subgraphs arent' actually useful when we condense to triples
    # we can add the predicates on the way back, the issue is that there
    # is a name collision with the prefixes that we would like to use
    # which is something of an annoying design failure I think ...
    #'subjects': 'TEMP:hasSubGraph',  #'_bfc:#subjects-graph',  # FIXME I think the fragment is the right thing to do here ...
    #'samples': 'TEMP:hasSubGraph',  #'_bfc:#samples-graph',
    #'contributors': '_bfc:#contributors-graph',  # FIXME broken I think
    #'contributors': 'TEMP:hasSubGraph',  # FIXME broken I think
    #'N:dataset': {'@id': 'https://api.pennsieve.io/datasets/'},  # darn it multiprefix issues
    'idfall': {"@id": str(TEMP['id-fallthrough/']), "@prefix": True},
    **{p: {'@id': n, '@prefix': True}
       if [c for c in prefix_endswith if n.endswith(c)]
       else n for p, n in {**uPREFIXES, **OntCuries._dict}.items()
       if p != ''},  # FIXME massive data duplication

    # OLD convention
    'unit': {'@id': 'TEMP:hasUnit',
             '@type': '@id',
             '@context': {'@base': str(unit)}},
    'value': {'@id': 'rdf:value'},  # FIXME maybe include this in a dedicated quantity @context?
    # END OLD

    'units': {'@id': 'TEMP:hasUnit',
             '@type': '@id',
             '@context': {'@base': str(unit)}},
    'magnitude': {'@id': 'rdf:value'},  # FIXME maybe include this in a dedicated quantity @context?
    'type': {'@id': 'ilxtr:jsonRecordType',  # FIXME ...
             '@type': '@id',
             '@context': {'@base': str(ilxtr['sparcur/types/'])},},
    'quantity': 'sparc:Measurement',  # for some reason this isn't working, and is captured by @base
    #'identifier': 'owl:Class',
    'system': {'@id': 'ilxtr:identifierType',  # FIXME see what predicate we used for this
               '@type': '@id',
               '@context': {'@base': str(ilxtr['idlib/systems/'])},},
    'OntTerm': 'owl:Class',
    #'uri_api': {'@id': '@id', '@type': '@id'},  # klobbered below, now fixed
}


protcur_context = {  # FIXME how is the protcur -> mis related to a native protcur schema?
    '@version': 1.1,
    #'TEMP:':

    '@base': 'https://uilx.org/tgbugs/u/sparcur-protcur-json-ld/',  # FIXME per subsection?

    'label': 'rdfs:label',

    'spjl': {'@id': 'https://uilx.org/tgbugs/u/sparcur-protcur-json-ld/', '@prefix': True},

    # simple page note prefixes
    'sparc-spn': {'@id': 'https://uilx.org/tgbugs/u/sparc/spn/', '@prefix': True},
    'anat': {'@id': 'https://uilx.org/tgbugs/u/sparc/spn/anatomy/', '@prefix': True},
    'apro': {'@id': 'https://uilx.org/tgbugs/u/sparc/spn/approach/', '@prefix': True},
    'tax': {'@id': 'https://uilx.org/tgbugs/u/sparc/spn/taxon/', '@prefix': True},

    # XXX FIXME annoying context collisions
    'unt': {'@id': str(unit), '@prefix': True},

    #'protc': {'@id': 'https://uilx.org/tgbugs/u/protc/', '@prefix': True},

    #'protcur': {'@id': 'https://uilx.org/tgbugs/u/protcur/', '@prefix': True},
    #'hyp-protcur': {'@id': 'https://uilx.org/tgbugs/u/hypothesis/protcur/', '@prefix': True},

    #'sparcur': {'@id': 'https://uilx.org/tgbugs/u/sparcur/', '@prefix': True},
    #'aspect': {'@id': 'https://uilx.org/tgbugs/u/aspect/', '@prefix': True},

    #'aspect-raw': {'@id': 'https://uilx.org/tgbugs/u/aspect-raw/', '@prefix': True},

    #'verb': {'@id': 'https://uilx.org/tgbugs/u/executor-verb/', '@prefix': True},

    'start': {'@id': 'TEMP:rangeStart', '@type': '@id'},
    'stop': {'@id': 'TEMP:rangeStop', '@type': '@id'},

    'hasDoi': {'@id': 'TEMP:hasDoi', '@type': '@id'},

    'fuzzy': {'@id': 'https://uilx.org/tgbugs/u/fuzzy-quantity/', '@prefix': True},

    'children': {'@id': 'TEMP:protcurChildren', '@type': '@id'},
    'ast_value': {'@id': 'TEMP:hasValue', '@type': '@id'},
    'raw_value': {'@id': 'TEMPRAW:hasValue', '@type': '@id'},
    'original': {'@id': 'TEMPRAW:hasValue'},  # don't want @type @id for this one

    'protcur_anno_count': {'@id': 'TEMP:hasNumberOfProtcurAnnotations'},
    'datasets': {'@id': 'TEMP:priorInformationalConstraintOnProcessThatGenerated', '@type': '@id'},

    'ilxtr_technique': {'@id': 'TEMPRAW:protocolEmploysTechnique'},
    'protocolEmploysTechnique': {'@id': 'TEMP:protocolEmploysTechnique', '@type': '@id'},

    # https://www.w3.org/TR/json-ld11/#example-32-using-vocabularies
    # explains the issues associated with trying to use a context to
    # to identifier remapping directly, you have to use an intermediate
    # context which compacts a form where keys are not curies at all
    'protc_aspect': {'@id': 'TEMPRAW:protocolInvolvesAspect'},
    'protc_implied-aspect': {'@id': 'TEMPRAW:protocolInvolvesAspect'},
    'protc_input': {'@id': 'TEMPRAW:protocolInvolvesInput'},
    'protc_implied-input': {'@id': 'TEMPRAW:protocolInvolvesInput'},
    'protc_input-instance': {'@id': 'TEMPRAW:protocolInvolvesInputInstance'},
    'protc_output': {'@id': 'TEMPRAW:protocolInvolvesOutput'},
    'protc_parameter*': {'@id': 'TEMPRAW:protocolInvolvesParameter'},
    'protc_invariant': {'@id': 'TEMPRAW:protocolInvolvesInvariant'},
    'protc_executor-verb': {'@id': 'TEMPRAW:protocolInvolvesAction'},
    'protc_black-box': {'@id': 'TEMPRAW:protocolInvolvesBlackBox'},
    'protc_black-box-component': {'@id': 'TEMPRAW:protocolInvolvesBlackBoxComponent'},
    'protc_objective*': {'@id': 'TEMPRAW:protocolInvolvesObjective'},

    'protcur_aspect': {'@id': 'TEMP:protocolInvolvesAspect', '@type': '@id'},
    'protcur_implied-aspect': {'@id': 'TEMP:protocolInvolvesAspect', '@type': '@id'},
    'protcur_input': {'@id': 'TEMP:protocolInvolvesInput', '@type': '@id'},
    'protcur_implied-input': {'@id': 'TEMP:protocolInvolvesInput', '@type': '@id'},
    'protcur_input-instance': {'@id': 'TEMP:protocolInvolvesInputInstance', '@type': '@id'},
    'protcur_output': {'@id': 'TEMP:protocolInvolvesOutput', '@type': '@id'},
    'protcur_parameter': {'@id': 'TEMP:protocolInvolvesParameter', '@type': '@id'},
    'protcur_invariant': {'@id': 'TEMP:protocolInvolvesInvariant', '@type': '@id'},
    'protcur_executor-verb': {'@id': 'TEMP:protocolInvolvesAction', '@type': '@id'},
    'protcur_black-box': {'@id': 'TEMP:protocolInvolvesBlackBox', '@type': '@id'},
    'protcur_black-box-component': {'@id': 'TEMP:protocolInvolvesBlackBoxComponent', '@type': '@id'},
    'protcur_objective': {'@id': 'TEMP:protocolInvolvesObjective', '@type': '@id'},

    # sparc simple page note curation workflow
    'sparc-anatomy': {'@id': 'sparc-spn:anatomy', '@type': '@id'},
    'sparc-approach': {'@id': 'sparc-spn:approach', '@type': '@id'},
    'sparc-taxon': {'@id': 'sparc-spn:taxon', '@type': '@id'},
    'sparc-other': {'@id': 'sparc-spn:other', '@type': '@id'},

    #TODO aspects of subjects vs reagents: {'@id': primary particiant in the
    #core protocols vs subprotocols

    'sparc_AnatomicalLocation': {'@id': 'TEMPRAW:involvesAnatomicalRegion'},
    'sparc_Measurement': {'@id': 'TEMPRAW:protocolMakesMeasurement'},
    'sparc_Reagent': {'@id': 'TEMPRAW:protocolUsesReagent'},
    'sparc_Tool': {'@id': 'TEMPRAW:protocolUsesTool'},
    'sparc_Sample': {'@id': 'TEMPRAW:protocolInvolvesSampleType'},
    'sparc_OrganismSubject': {'@id': 'TEMPRAW:protocolInvolvesSubjectType'},
    'sparc_Procedure': {'@id': 'TEMPRAW:whatIsThisDoingHere'},

    'sparcur_AnatomicalLocation': {'@id': 'TEMP:involvesAnatomicalRegion', '@type': '@id'},
    'sparcur_Measurement': {'@id': 'TEMP:protocolMakesMeasurement', '@type': '@id'},
    'sparcur_Reagent': {'@id': 'TEMP:protocolUsesReagent', '@type': '@id'},
    'sparcur_Tool': {'@id': 'TEMP:protocolUsesTool', '@type': '@id'},
    'sparcur_Sample': {'@id': 'TEMP:protocolInvolvesSampleType', '@type': '@id'},
    'sparcur_OrganismSubject': {'@id': 'TEMP:protocolInvolvesSubjectType', '@type': '@id'},
    'sparcur_Procedure': {'@id': 'TEMP:whatIsThisDoingHere', '@type': '@id'},
}


def json_version(version):
    """ version should be a branch or a release tag e.g. schema-0.0.1 """
    return {
        '$schema': 'http://json-schema.org/draft-06/schema#',
        '$id': ('https://raw.githubusercontent.com/SciCrunch/sparc-curation/'
                f'{version}/resources/defs.schema.json'),
        'title': 'Common definitions for the SDS json schemas, including mappings to json-ld @context',
        'description': 'JSON-LD mappings are under the `context_value` key by convention',
        '@context': {
            'sparc': str(sparc),
            'TEMP': str(TEMP),
            'TEMPRAW': str(TEMPRAW),
        },
        'definitions': {
            'patterns': {
                'metadata_filename': metadata_filename_pattern,
                'simple_url': simple_url_pattern ,
                #'doi': idlib.Doi._id_class.canonical_regex,
                #'orcid': idlib.Orcid._id_class.canonical_regex,
                #'ror': idlib.Ror._id_class.canonical_regex,
            'iso8601_date': iso8601datepattern,
                'iso8601_datetime': iso8601pattern,
                'whitespace_lead_trail': pattern_whitespace_lead_trail,
                'whitespace_multi': pattern_whitespace_multi,
            },
            'NoLTWhitespace': NoLTWhitespaceSchema.schema,
            'EmbeddedIdentifier': EmbeddedIdentifierSchema.schema,

            'meta': MetaOutExportSchema.schema,
            'prov': ProvSchema.schema,
            'errors': ErrorSchema.schema,
            'contributor': ContributorOutExportSchema.schema,
            'performance': PerformanceExportSchema.schema,
            'subject': SubjectExportSchema.schema,
            'sample': SampleExportSchema.schema,
            'dataset_description_file': DatasetDescriptionSchema.schema,
            'submission_file': SubmissionSchema.schema,
            'subjects_file': SubjectsSchema.schema,
            'samples_file': SamplesFileSchema.schema,
            'dataset': {
                'meta': {'$ref': '#/definitions/meta'},
                'prov': {'$ref': '#/definitions/prov'},
                'errors': {'$ref': '#/definitions/errors'},
                'contributor': {'$ref': '#/definitions/contributor'},
                'subject': {'$ref': '#/definitions/subject'},
                'sample': {'$ref': '#/definitions/sample'},
                'dataaset_description_file': {'$ref': '#/definitions/dataaset_description_file'},
                'submission_file': {'$ref': '#/definitions/submission_file'},
                'subjects_file': {'$ref': '#/definitions/subjects_file'},
                'samples_file': {'$ref': '#/definitions/samples_file'},
                'errors': {'$ref': '#/definitions/errors'},
            }
        }
    }


def get_runtime_context_specs():
    # FIXME not clear we should be embedding context_runtime deeply in
    # the schemas, though it is useful for being able to see things
    # side by side
    from pyontutils.utils import subclasses
    collect = []
    for cls in subclasses(JSONSchema):
        JApplyRecursive(get_nested_by_key,
                        cls.schema,
                        'context_runtime',
                        join_lists=False,
                        collect=collect)

    specs = tuple(set(tuple(l) for l in collect))
    # FIXME convert to JPointer here?
    return specs


def _defunc(obj, *args, path=None, **kwargs):  # FIXME add tests
    if isinstance(obj, types.FunctionType):
        return asStr(ast.parse(inspect.getsource(obj).strip().strip(',')))
    else:
        return obj


def update_include_paths(obj, *args, path=None, key='jsonld_include', moves=tuple(), collect=tuple()):
    """ collect json pointers that need to be updated """
    if isinstance(obj, dict) and key in obj:
        path = list(JPointer.pathFromSchema(path))
        for frm, to in moves:  # sigh python
            lf = len(frm)
            if path[:lf] == frm:
                path = to + path[lf:]
                break

        for key, value in obj[key].items():
            path.append(key)
            collect.append((path, lambda _, _v=value: _v))  # note that collect is really a list
            # note also that lambda _: value is because pipeline adds currently calls into lifters

    return obj  # have to return this otherwise somehow everything is turned to None?


def idtype(id, base=str(TEMP['id-fallthrough/'])):
    # we set base to id-fallthrough so we can easily pull out unmapped
    # cases that should be identifiers
    out = {'@id': id, '@type': '@id'}
    if base and id != '@id':
        # for whatever reason @context for @id @id is not allowed so
        # it should be set in an outer @context
        out['@context'] = {'@base': str(base)}

    return out


def strcont(id):
    return {'type': 'string', 'context_value': id}


def inttype(id):
    return {'@id': id, '@type': 'xsd:integer'}


def booltype(id):
    return {'@id': id, '@type': 'xsd:boolean'}


def cont(id, type):
    return {'type': type, 'context_value': id}


def not_array(schema, in_all=False):
    """ hacked way to determine that the blob in question cannot be an array """
    return ('type' in schema and schema['type'] != 'array' or
            'properties' in schema and 'items' not in schema or
            # FIXME hack hard to maintain
            'allOf' in schema and all(not_array(s, True) for s in schema['allOf']) or
            'oneOf' in schema and all(not_array(s, True) for s in schema['oneOf']) or
            'anyOf' in schema and all(not_array(s, True) for s in schema['anyOf']) or
            'not' in schema and (not not_array(schema['not'], in_all) or in_all)
    )


class hproperty:
    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        self.fget = fget
        self.fset = fset
        self.fdel = fdel
        if doc is None and fget is not None:
            doc = fget.__doc__
        self.__doc__ = doc

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self

        if self.fget is None:
            raise AttributeError("unreadable attribute")

        return self.fget(obj)

    def __set__(self, obj, value):
        if self.fset is None:
            raise AttributeError("can't set attribute")

        self.fset(obj, value)

    def __delete__(self, obj):
        if self.fdel is None:
            raise AttributeError("can't delete attribute")

        self.fdel(obj)

    def getter(self, fget):
        return type(self)(fget, self.fset, self.fdel, self.__doc__)

    def setter(self, fset):
        return type(self)(self.fget, fset, self.fdel, self.__doc__)

    def deleter(self, fdel):
        return type(self)(self.fget, self.fset, fdel, self.__doc__)


class HasSchema:
    """ decorator for classes with methods whose output can be validated by jsonschema """
    def __init__(self, input_schema_class=None, fail=True, normalize=False):
        self.input_schema_class = input_schema_class
        self.fail = fail
        self.normalize = normalize
        self.schema = None  # deprecated output schema ...

    def mark(self, cls):
        """ note that this runs AFTER all the methods """

        if self.input_schema_class is not None:
            # FIXME probably better to do this as types
            # and check that schemas match at class time
            cls._pipeline_start = cls.pipeline_start

            @hproperty
            def pipeline_start(self, isc=self.input_schema_class, fail=self.fail):
                if pipeline_start.schema == isc:
                    pipeline_start.schema = isc()

                schema = pipeline_start.schema
                data = self._pipeline_start
                ok, norm_or_error, data = schema.validate(data)
                if not ok and fail:
                    raise norm_or_error

                return data

            pipeline_start.schema = self.input_schema_class
            cls.pipeline_start = pipeline_start

        return cls

    def f(self, schema_class, fail=False):
        return self(schema_class, fail=fail, _property=False)

    def __call__(self, schema_class, fail=False, _property=True):
        # TODO switch for normalized output if value passes?
        def decorator(function):
            pipeline_stage_name = function.__qualname__
            @wraps(function)
            def schema_wrapped_property(_self, *args, **kwargs):
                if schema_wrapped_property.schema == schema_class:
                    schema_wrapped_property.schema = schema_class()

                schema = schema_wrapped_property.schema
                data = function(_self, *args, **kwargs)
                ok, norm_or_error, data = schema.validate(data)
                if not ok:
                    if fail:
                        logd.error('schema validation has failed and fail=True')
                        logd.critical(f'failing dataset is {data["id"]}')
                        raise norm_or_error

                    # I have no idea why this if statement used to be in
                    # a try except block, it should never fail unless
                    # the definition of __contains__ for a dict was overwritten
                    if 'errors' not in data:
                        data['errors'] = []
                        
                    data['errors'] += norm_or_error.json(pipeline_stage_name)
                    # TODO make sure the step is noted even if the schema is the same

                if self.normalize:
                    return norm_or_error

                return data

            if _property:
                schema_wrapped_property = hproperty(schema_wrapped_property)

            schema_wrapped_property.schema = schema_class
            return schema_wrapped_property

        return decorator


class ConvertingChecker(jsonschema.FormatChecker):
    _enc = JEncode()

    def check(self, instance, format):
        converted = _enc.default(instance)
        return super().check(converted, format)


def _make_st_checker(st):
    def inn(c, i):
        for t in st:
            if isinstance(i, t):
                return True

    return inn


class JSONSchema(object):

    schema = {}

    string_types = [  # add more types here
        str,
        PurePath,
        idlib.Stream,  # FIXME str/dict
        rdflib.URIRef,
        rdflib.Literal,
        oq.OntId,
        oq.OntTerm,  # FIXME str/dict
        ProtcurExpression,
    ]

    type_checker = jsonschema.Draft6Validator.TYPE_CHECKER.redefine_many(
        dict(array=(lambda c, i: isinstance(i, list) or isinstance(i, tuple)),
             string=_make_st_checker(string_types)))

    validator_class = jsonschema.validators.extend(
        jsonschema.Draft6Validator,
        type_checker=type_checker)

    def __init__(self):
        format_checker = jsonschema.FormatChecker()
        #format_checker = ConvertingChecker()
        types = dict(array=(list, tuple),
                     string=tuple(self.string_types))
        self.validator = self.validator_class(self.schema,
                                              format_checker=format_checker)

    @classmethod
    def _add_meta(cls, blob):
        """ In place modification of a schema blob to add schema level metadata """
        # TODO schema metadata, I don't know enough to do this correctly yet
        #template_schema_version = cls.validator_class.ID_OF(cls.validator_class.META_SCHEMA)
        #schema['$schema'] = template_schema_version

        #schema_id = '#/temp/' + cls.__name__ + '.json'  # FIXME version ...
        #schema['$id'] = schema_id

    @classmethod
    def export(cls, base_path):
        """ write schema to file with """
        def pop_errors(dict_):
            if 'errors' in dict_:
                dict_.pop('errors')
            for v in dict_.values():
                if isinstance(v, dict):
                    pop_errors(v)

        schema = copy.deepcopy(cls.schema)
        pop_errors(schema)
        cls._add_meta(schema)

        def debug(obj, *args, path=None, **kwargs):
            if isinstance(obj, types.FunctionType):
                raise TypeError(path + [obj])
            else:
                return obj

        schema = JApplyRecursive(_defunc, schema)
        JApplyRecursive(debug, schema)

        with open((base_path / cls.__name__).with_suffix('.json'), 'wt') as f:
            json.dump(schema, f, sort_keys=True, indent=2)

    def validate_strict(self, data):
        # Take a copy to ensure we don't modify what we were passed.
        #appstruct = copy.deepcopy(data)

        appstruct = json.loads(json.dumps(data, cls=JEncode))  # FIXME figure out converters ...

        errors = list(self.validator.iter_errors(appstruct))
        if errors:
            raise exc.ValidationError(errors)

        return appstruct

    def validate(self, data):
        """ capture errors """
        try:
            ok = self.validate_strict(data)  # validate {} to get better error messages
            return True, ok, data  # FIXME better format

        except exc.ValidationError as e:
            return False, e, data  # FIXME better format

    @classmethod
    def context(cls):
        """ return a json-ld context by extracting annotations from nested schemas
            those annotations are currently placed under the `context_value' key """
        # FIXME last one wins right now
        #context_runtimes = {}
        context = {}
        marker = 'context_value'
        #marker_runtime = 'context_runtime'
        def inner(schema, key=None):
            if isinstance(schema, dict):
                if marker in schema:
                    if key is None:
                        raise ValueError('you probably don\'t want to do this ...')

                    context_value = schema[marker]
                    context[key] = context_value

                #if marker_runtime in schema:
                    #if key is None:
                        #raise ValueError('you probably don\'t want to do this ...')

                    #context_runtimes[key] = schema[marker_runtime]

                for k, v in schema.items():
                    if k == marker:
                        continue
                    inner(v, k)

            elif isinstance(schema, list):
                for ss in schema:
                    inner(ss, key)

            else:
                pass

        inner(cls.schema)

        #if context:  # if there is no context do not add version
            #context.update(base_context)

        #if context_runtimes:
            #def context_runtime(base):
                #return {k: v(base) for k, v in context_runtimes.items()}

        return context


class JsonLdHelperSchema(JSONSchema):
    """ documentation of the additions to the json schema draft validator schema
        that are used for transformation to jsonld """
    schema = {
        'properties': {
            'jsonld_includes': {'type': 'object'},
            'context_value': {'oneOf': [
                # re https://github.com/json-ld/json-ld.org/issues/612
                # https://github.com/json-ld/json-ld.org/blob/master/schemas/jsonld-schema.json
                {'type': 'string'},  # TODO are there any restrictions? string -> @base/string ?
                {'type': 'object'},  # TODO this should conform to the jsonld schema for @context
            ]}
        }
    }


class RuntimeSchema(JSONSchema):
    """ finally a zero (runtime) cost first time only init
    with reasonable debugability and readability """

    schema = None

    def __new__(cls):
        return super().__new__(cls)

    _renew = __new__

    def __new__(cls, *args, **kwargs):
        cls.setup(*args, **kwargs)
        cls.__new__ = cls._renew
        return super().__new__(cls)

    def __init__(self):
        super().__init__()

    _reinit = __init__

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.__class__.__init__ = self.__class__._reinit

    @classmethod
    def setup(cls):
        raise NotImplementedError('subclassit')


class RemoteSchema(RuntimeSchema):
    schema = ''
    # FIXME TODO these need to be cached locally
    @classmethod
    def setup(cls):
        cls.schema = requests.get(cls.schema).json()


class ApiNATOMYSchema(RemoteSchema):
    schema = ('https://raw.githubusercontent.com/open-physiology/'
              'open-physiology-viewer/master/src/model/graphScheme.json')


class ToExport(RuntimeSchema):
    """ version should match the github release tag i.e. schema-0.0.1 """

    @classmethod
    def setup(cls, version=None):
        if version is None:
            version = 'master'
        #cls.schema = JApplyRecursive(_defunc, json_version(version))
        cls.schema = json_version(version)


metadata_filename_pattern = r'^.+\/[A-Za-z-_\/]+\.(xlsx|csv|tsv|json)$'

simple_url_pattern = r'^(https?):\/\/([^\s\/]+)\/([^\s]*)'
simple_file_url_pattern = r'^file:\/\/'

# no whitespace no colons for subject and sample identifiers
# and no forward slashes and no pipes
# also no leading symbols (problematic for utf-8 but deal with it)
# see https://stackoverflow.com/questions/1976007/
fs_safe_identifier_pattern = '^[A-Za-z0-9][^ \n\t*|:\/<>"\\?]+$'

# NOTE don't use builtin date-time format due to , vs . issue
iso8601pattern =     '^[0-9]{4}-[0-1][0-9]-[0-3][0-9]T[0-2][0-9]:[0-6][0-9](:[0-6][0-9](,[0-9]{1,9})?)?(Z|[+-][0-2][0-9]:[0-6][0-9])$'
iso8601bothpattern = '^[0-9]{4}-[0-1][0-9]-[0-3][0-9](T[0-2][0-9]:[0-6][0-9](:[0-6][0-9](,[0-9]{1,9})?)?(Z|[+-][0-2][0-9]:[0-6][0-9]))?$'
iso8601datepattern = '^[0-9]{4}-[0-1][0-9]-[0-3][0-9]$'  # XXX people are using this in the timestamp field repeatedly

# https://www.crossref.org/blog/dois-and-matching-regular-expressions/
doi_pattern = idlib.Doi._id_class.canonical_regex

orcid_pattern = idlib.Orcid._id_class.canonical_regex

# for a fun (?) time
# https://en.wikipedia.org/wiki/List_of_family_name_affixes
contributor_name_pattern = r'^((([Vv](an|on))|[Dd][ia]|([Dd]e( [Ll]os)?)) )?[^, ]+, [^,]+$'

ror_pattern = idlib.Ror._id_class.canonical_regex

pattern_whitespace_lead_trail = r'(^[\s]+[^\s].*|.*[^\s][\s]+$)'
pattern_whitespace_multi = r'[\s]{2,}'

award_pattern = '^(OT2OD|OT3OD|U18|TR|U01)'  # really just no whitespace or / ?


class NoLTWhitespaceSchema(JSONSchema):
    schema = {'allOf': [{'type': 'string'},
                        {'not': {'pattern': pattern_whitespace_lead_trail,}},
                        {'not': {'pattern': pattern_whitespace_multi}},
                        ]}


string_noltws = NoLTWhitespaceSchema.schema


class ErrorSchema(JSONSchema):
    schema = {'type':'array',
              'minItems': 1,
              'items': {'type': 'object'},}



def make_id_schema(cls, pattern=None, format=None):
    id_schema = {'type': 'string'}
    if format is not None:
        id_schema['format'] = format

    if pattern is not None:
        id_schema['pattern'] = pattern
    elif hasattr(cls, '_id_class') and hasattr(cls._id_class, 'canonical_regex'):
        id_schema['pattern'] = cls._id_class.canonical_regex

    class _IdSchema(JSONSchema):
        __name__ = cls.__name__ + 'Schema'
        schema = {'properties': {'system': {'type': 'string',
                                            'enum': [cls.__name__],},
                                 'id': id_schema},}

    return _IdSchema


OntTermSchema = make_id_schema(oq.OntTerm, format='iri')
DoiSchema = make_id_schema(idlib.Doi)
RorSchema = make_id_schema(idlib.Ror)
OrcidSchema = make_id_schema(idlib.Orcid)
PioSchema = make_id_schema(idlib.Pio)
RridSchema = make_id_schema(idlib.Rrid)
BlackfynnIdSchema = make_id_schema(BlackfynnId)
PennsieveIdSchema = make_id_schema(PennsieveId)


class EmbeddedIdentifierSchema(JSONSchema):
    schema = {'type': 'object',
              'required': ['type', 'id', 'label'],
              'properties': {
                  'id': {'type': 'string'},
                  'type': {'type': 'string',
                           'enum': ['identifier']},
                  'system': {'type': 'string'},
                  'label': {'type': 'string'},
                  'synonyms': {'type': 'array',
                               'minItems': 1},
                  # NOTE: description is the normalized form name ofr
                  # ontology term definitions
                  'description': {'type': 'string'},},}

    @classmethod
    def _to_pattern(cls, obj, *args, path=None, alternates=None, **kwargs):  # FIXME add tests
        if (isinstance(obj, dict) and
            'allOf' in obj and
            (len(obj) == 1 or len(obj) == 2 and 'context_value' in obj) and  # FIXME extremely brittle
            len(obj['allOf']) == 2 and
            obj['allOf'][0] == cls.schema):
            pattern = obj['allOf'][1]['properties']['id']
            if alternates is not None:
                # FIXME SUPER brittle
                system = obj['allOf'][1]['properties']['system']['enum'][0]
                if system in alternates:
                    pattern['pattern'] = pattern['pattern'] + '|' + alternates[system]

            return pattern

        else:
            return obj

    @classmethod
    def _allOf(cls, schema, context_value=None):
        ao = {'allOf': [cls.schema, schema.schema,]}
        if context_value:
            ao['context_value'] = context_value

        return ao

    @classmethod
    def allOf(cls, schema_):
        class _SCM(JSONSchema):
            schema = cls._allOf(schema_)  # LOL PYTHON

        return _SCM


EIS = EmbeddedIdentifierSchema
EISs = EmbeddedIdentifierSchema.schema


class ProvSchema(JSONSchema):
    schema = {'type': 'object',
              'properties': {'timestamp_export_start': {'type': 'string',
                                                        'pattern': iso8601pattern,},
                             #'timestamp_export_end': {'type': 'string'},  # this isn't really possible
                             'export_system_identifier': {'type': 'string'},
                             'export_system_hostname': {'type': 'string'},
                             'export_project_path': {'type': 'string'},
                             'remote': {'type': 'string'},
                             'sparcur_version': {'type': 'string'},
                             'sparcur_internal_version': {'type': 'integer'},
                             'sparcur_commit': {'type': 'string'},
                             'ontology_mappings': {
                                 'type': 'array',
                                 'minItems': 1,
                                 'items': {
                                     'type': 'object',
                                     'properties': {
                                         'curie': {'type': 'string'},
                                         'label': {'type': 'string'},
                                         'matches': {
                                             'type': 'array',
                                             'minItems': 1,
                                             'items': {
                                                 'type': 'object',
                                                 'properties': {
                                                     'path': {'type': 'array'},
                                                     'input': {'type': 'string'},
                                                 },},},},},},},}


class MbfTracingSchema(JSONSchema):
    schema = {
        'type': 'object',
        'required': ['subject', 'atlas', 'images', 'contours'],
        'properties': {
            'subject': {'type': 'object',
                        'required': ['id'],
                        'properties': {'id': {'type': 'string'},
                                       'species': {'type': 'string'},
                                       'sex': {'type': 'string'},
                                       'age': {'type': 'string'},
                                       },},
            'atlas': {'type': 'object',
                      'properties': {'organ': {'type': 'string'},
                                     'atlas_label': {'type': 'string'},
                                     'atlas_rootid': {'type': 'string'},
                                     },},
            'images': {'type': 'array',
                       'items': {'type': 'object',
                                 'properties':
                                 {'path_mbf': {
                                     'type': 'array',
                                     'minItems': 1,
                                     'items': {'type': 'string'}},
                                  'channels': {
                                      'type': 'array',
                                      'items': {
                                          'type': 'object',
                                          'properties': {
                                              'id': {'type': 'string'},
                                              'source': {'type': 'string'},
                                          },},},},},},
            'contours': {'type': 'array',
                         'items': {'type': 'object',
                                   'properties': {
                                       'name': string_noltws,
                                       'guid': {'type': 'string'},
                                       'id_ontology': {'type': 'string'},},
                                   },},},}


class NeurolucidaSchema(JSONSchema):
    schema = copy.deepcopy(MbfTracingSchema.schema)
    schema['required'] = ['images', 'contours']


class DatasetStructureSchema(JSONSchema):
    __schema = {'type': 'object',
                'required': ['submission_file', 'dataset_description_file'],
                'properties': {
                    'submission_file': {'type': 'string',
                                        'pattern': metadata_filename_pattern},
                    'dataset_description_file': {'type': 'string',
                                                 'pattern': metadata_filename_pattern},
                    'subjects_file': {'type': 'string',
                                      'pattern': metadata_filename_pattern},
                    'samples_file': {'type': 'string',
                                     'pattern': metadata_filename_pattern},
                    'manifest_file': {'type': 'array',
                                      'minItems': 1,
                                      'items': {
                                          'type': 'string',
                                          'pattern': metadata_filename_pattern},
                                      },
                    'dir_structure': {'type': 'array',  # TODO FIXME do not include in primary export
                                      'items': {'type': 'object',}
                                      ,},
                    'dirs': {'type': 'integer',
                             'minimum': 0},
                    'files': {'type': 'integer',
                              'minimum': 0},
                    'size': {'type': 'integer',
                             'minimum': 0},
                    'errors': ErrorSchema.schema,
                }}
    schema = {'allOf': [__schema,
                        {'anyOf': [
                            {'required': ['subjects_file']},
                            {'required': ['samples_file']}]}]}


class FutureValidPaths:
    # maybe we use these in the future?

    folder_patterns_full = (
        '^('
        'part(icipant)?|'
        'pop(ulation)?|'
        'spec(imen)?|'
        'sub(j|ject)?|'
        'sam(p|ple)?|'
        'pool|'
        'perf(ormance)?'
        ')-([0-9A-Za-z][0-9A-Za-z-]*)$'
    )

    folder_patterns = '^(part|pop|spec|subj?|samp?|pool|perf)-([0-9A-Za-z][0-9A-Za-z-]*)$'
    folder_patterns_strict = '^(pop|sub|sam|pool|perf)-([0-9A-Za-z][0-9A-Za-z-]*)$'

    participant_patterns = '^(part|pop|spec|subj?|samp?|pool)-([0-9A-Za-z][0-9A-Za-z-]*)$'
    participant_patterns_strict = '^(pop|sub|sam|pool)-([0-9A-Za-z][0-9A-Za-z-]*)$'

    ['0-9A-Za-z-']
    ['^pool-']
    ['^subj?-']
    ['^samp?-']
    ['^perf-']  # ^ses- ^scan- ^run- roll up here
    ['^spec-']
    ['^pop-']
    ['^part-']


    ['code', '*']
    ['docs', '*']
    ['source', '*']  # but will check for matches and warn on extra
    ['derivative', '*']  # but will check for matches and warn on extra

    # paths
    ['primary', '^pop-']
    ['primary', '^pool-']
    ['primary', '^pool-', '^perf-']
    ['primary', '^subj?-']
    ['primary', '^subj?-', '^perf-']
    ['primary', '^subj?-', '^samp?-']
    ['primary', '^samp?-']
    ['primary', '^samp?-', '^perf-']
    # ['primary', '^perf-']  # these would be perf types which are metadata so we don't allow this pattern

    # data modalities ^mod- in bids
    '^phys(iology)?$'
    '^seq(uencing)?$'
    '^rad(iology)?$'
    '^mic(roscopy)?$'
    '^morph(ology)?$'  # swc

    # XXX we might try to avoid using modalities
    # we are going to use ^perf-type-sub-1 instead or something like that
    # then the protocol should give us the data type, keep the data types out of this
    # OR do we need that information to statically validate allowed file types?

    # the more likely approach is as follows
    # all some-* indicates that the id MUST be a known
    # identifier whose type matches the expected slot

    # primary (or some-subject some-sample some-pool some-perf)
    # this is broken unless the manifest can save us (which it must)
    # otherwise we have to warn on all folders that do not match
    # a known entity
    # a perf- folder may appear once per non-perf node XXX this might be incorrect?
    # some-subject (or some-sample some-perf-no-subject some-manifest-folder)
    # some-sample (or some-sample some-perf-end some-manifest-folder)
    # some-perf-no-subject (or some-sample some-manifest-folder)
    # some-perf (or some-subject-no-perf some-sample-no-perf)
    # HOWEVER this is too much for right now, we may need to completely
    # rework the requirements for the template 2.0 to avoid these tricky cases
    # for now, we are going to find all folders that match an id below the top
    # level and then we are going to warn about all other folder names that are
    # not either contained in a subject where there are no samples or in a sample

    # {'allOf': [
    #     {'type': 'string',},
    #                   {'oneOf': [
    #                       {'pattern': subject_folder_pattern},
    #                       {'pattern': sample_folder_pattern},
    #                       {'pattern': performance_folder_pattern},
    #                       {'pattern': pool_folder_pattern},
    #                   ],},]
    # }


class DatasetPathSchema(JSONSchema):
    schema = {'type': 'array',
              'additionalItems': {'type': 'string'},
              'items': [
                  {'type': 'string',
                   'enum': ['code', 'docs', 'source', 'primary', 'derivative'],
                   },]}


class ContributorExportSchema(JSONSchema):
    schema = {
        'type': 'object',
        #'jsonld_include': {'@type': ['sparc:Person', 'owl:NamedIndividual']},  # FIXME being included THREE times due to references in nested schema # XXX THIS HAS BEEN MOVED AFTER ALL DEEPCOPY OPERATIONS SIGH
        'properties': {
            'id': {'type': 'string',  # FIXME TODO
                   },
            'affiliation': EIS._allOf(RorSchema,
                                      context_value=idtype('TEMP:hasAffiliation')),
            'contributor_name': {'type': 'string',
                                 'pattern': contributor_name_pattern,},
            'first_name': {'type': 'string',
                           'context_value': 'sparc:firstName',},
            'middle_name': strcont('TEMP:middleName'),
            'last_name': {'type': 'string',
                          'context_value': 'sparc:lastName',},
            'contributor_orcid': EIS._allOf(
                OrcidSchema,
                context_value=idtype('sparc:hasORCIDId')),
            #'data_remote_user_id': EIS._allOf(
                #PennsieveIdSchema,
                #context_value=idtype('TEMP:hasDataRemoteUserId')),
            'data_remote_user_id': strcont(idtype('TEMP:hasDataRemoteUserId')),
            #'blackfynn_user_id': EIS._allOf(
                #BlackfynnIdSchema,
                #context_value=idtype('TEMP:hasBlackfynnUserId')),
            'blackfynn_user_id': strcont(idtype('TEMP:hasBlackfynnUserId')),
            'contributor_affiliation': {
                'anyOf': [EIS._allOf(RorSchema),  # FIXME oneOf issues with EIS
                          {'type': 'string'}],
                'context_value': idtype('TEMP:hasAffiliation')
            },
            'contributor_role': {
                'type':'array',
                'context_value': idtype('TEMP:hasRole', base=str(TEMP)),
                'minItems': 1,
                'items': {
                    'type': 'string',
                    'enum': ['CorrespondingAuthor',  # FIXME max card 1
                             'ContactPerson',  # FIXME need version dependence here
                             'Creator',  # allowed here, moved later
                             'DataCollector',
                             'DataCurator',
                             'DataManager', # this is DatasetMaintainer
                             'Distributor',
                             'Editor',
                             'HostingInstitution',
                             'PrincipalInvestigator',  # added for sparc map to ProjectLeader probably?
                             'CoInvestigator', # added for sparc, to distingusih ResponsibleInvestigator
                             'Producer',
                             'ProjectLeader',
                             'ProjectManager',
                             'ProjectMember',
                             'RegistrationAgency',
                             'RegistrationAuthority',
                             'RelatedPerson',
                             'Researcher',
                             'ResearchGroup',
                             'RightsHolder',
                             'Sponsor',
                             'Supervisor',
                             'WorkPackageLeader',
                             'Other',]}
            },
            'is_contact_person': {'type': 'boolean'},  # XXX deprecated
            'errors': ErrorSchema.schema,
        }}


class ContributorSchema(JSONSchema):
    context = lambda : ({}, None)
    __schema = copy.deepcopy(ContributorExportSchema.schema)
    schema = JApplyRecursive(
        EIS._to_pattern, __schema,
        alternates={'Orcid': idlib.Orcid._id_class.local_regex})


class ContributorsExportSchema(JSONSchema):
    schema = {'anyOf':
              [{'type': 'array',
                #'context_runtime': lambda base: {'@id': f'{base}contributors',},
                'contains': {
                    'type': 'object',
                    'required': ['is_contact_person'],
                    'properties': {
                        'is_contact_person': {'type': 'boolean', 'enum': [True]},
                    },
                },
                'items': ContributorExportSchema.schema,
                },
               # 2.0.0
               {'type': 'array',
                'contains': {
                    'type': 'object',
                    'required': ['contributor_role'],
                    'properties': {
                        'contributor_role': {
                            'type': 'array',
                            'contains': {
                                'type': 'string',
                                'enum': ['CorrespondingAuthor', 'ContactPerson']
                            },
                        },
                    },
                },
                'items': ContributorExportSchema.schema,},
               ],
              }


class ContributorsSchema(JSONSchema):
    context = lambda : ({}, None)
    __schema = copy.deepcopy(ContributorsExportSchema.schema)
    schema = JApplyRecursive(
        EIS._to_pattern, __schema,  # watch out for singular vs plural
        # >_< also watch out for non-commutative derivation chains,
        # sigh this is exactly the kind of issue that I was worried
        # about when I originally took this approach and yes, it is a
        # mistake and is why DRY is a good thing
        alternates={'Orcid': idlib.Orcid._id_class.local_regex})


class ContributorOutExportSchema(JSONSchema):
    schema = copy.deepcopy(ContributorExportSchema.schema)
    schema['properties']['contributor_role']['items']['enum'].remove('Creator')


class ContributorsOutExportSchema(JSONSchema):
    schema = copy.deepcopy(ContributorsExportSchema.schema)
    schema['items'] = ContributorOutExportSchema.schema


class CreatorExportSchema(JSONSchema):
    schema = copy.deepcopy(ContributorExportSchema.schema)
    schema['properties'].pop('contributor_role')
    schema['properties'].pop('is_contact_person')


class CreatorSchema(JSONSchema):
    context = lambda : ({}, None)
    __schema = copy.deepcopy(CreatorExportSchema.schema)
    schema = JApplyRecursive(EIS._to_pattern, __schema)



class CreatorsExportSchema(JSONSchema):
    schema = {'type': 'array',
              'minItems': 1,
              'items': CreatorExportSchema.schema}


class CreatorsSchema(JSONSchema):
    context = lambda : ({}, None)
    __schema = copy.deepcopy(CreatorsExportSchema.schema)
    schema = JApplyRecursive(EIS._to_pattern, __schema)


_protocol_url_or_doi_schema = {'anyOf':[EIS._allOf(DoiSchema),
                                        EIS._allOf(PioSchema),
                                        {'type': 'string',
                                         'pattern': simple_url_pattern,}]}


ContributorExportSchema.schema['jsonld_include'] = {
    '@type': ['sparc:Person', 'owl:NamedIndividual']}


class DatasetDescriptionExportSchema(JSONSchema):
    schema = {
        'type': 'object',
        'additionalProperties': False,
        'required': ['template_schema_version',  # missing should fail for this one ...
                     'name',
                     'description',
                     'funding',
                     'protocol_url_or_doi',
                     #'completeness_of_data_set'  # TODO versioned schema validation
                     'contributors',
                     'number_of_subjects',
                     'number_of_samples',],
        # TODO dependency to have title for complete if completeness is is not complete?
        'properties': {
            # 2.x helpers
            'dataset_type': {'type': 'string'},
            # 1.x
            'errors': ErrorSchema.schema,
            'template_schema_version': {'type': 'string'},
            'name': string_noltws,
            'description': {'type': 'string'},
            'keywords': {'type': 'array', 'items': {'type': 'string'}},
            'acknowledgments': {'type': 'string'},
            'funding': {'type': 'array',
                        'minItems': 1,
                        'items': {'type': 'string'}},
            'completeness_of_data_set': {'type': 'string'},
            'prior_batch_number': {'type': 'string'},
            'title_for_complete_data_set': {'type': 'string'},
            'originating_article_doi': {'type': 'array',
                                        'items': EIS._allOf(DoiSchema),
                                        #'context_value': idtype('TEMP:hasDoi'),  # FIXME convert here?
                                        },
            'number_of_subjects': {'type': 'integer'},
            'number_of_samples': {'type': 'integer'},
            'parent_dataset_id': {'type': 'string'},  # blackfynn id
            'protocol_url_or_doi': {'type': 'array',
                                    'minItems': 1,
                                    'items': _protocol_url_or_doi_schema,
                                    #'context_value': idtype('TEMP:hasProtocol'),  # FIXME convert here?
                                    },
            'links': {
                'type': 'array',
                'minItems': 1,
                'items': {
                    'type': 'object',
                    'properties': {
                        'additional_links': {'type': 'string'},
                        'link_description': {'type': 'string'},  # FIXME only if there is an additional link? some using for protocol desc
                    }
                }
            },
            'related_identifiers' : {
                'type': 'array',
                'items': {
                    'type': 'object',
                    'required': [
                        'relation_type',
                        'related_identifier_type',
                        'related_identifier'],
                    'properties': {
                        'related_identifier': {'type': 'string'},
                        'related_identifier_type': {
                            'type': 'string',
                            'enum': [
                                # datacite types
                                'ARK',
                                'arXiv',
                                'bibcode',
                                'DOI',
                                'EAN13',
                                'EISSN',
                                'Handle',
                                'IGSN',
                                'ISBN',
                                'ISSN',
                                'ISTC',
                                'LISSN',
                                'LSID',
                                'PMID',
                                'PURL',
                                'UPC',
                                'URL',
                                'URN',
                                'w3id',
                            ],},
                        'relation_type': {
                            'type': 'string',
                            'enum':
                            [
                                # additional relations that are needed for sparc, specifically for code and protocols
                                'IsProtocolFor',  # like is metadata for
                                'HasProtocol', # hopefully entails WasGeneratedByFollowing
                                'IsSoftwareFor', # includes source code
                                'HasSoftware',
                                # DataCite conflates source and compiled, consider Requires?
                                # Dataset -> Software -> SourceCode

                                # datacite relations as of schema 4.4
                                'IsCitedBy',
                                'Cites',
                                'IsSupplementTo',
                                'IsSupplementedBy',
                                'IsContinuedByContinues',
                                'IsDescribedBy',
                                'Describes',
                                'HasMetadata',
                                'IsMetadataFor',
                                'HasVersion',
                                'IsVersionOf',
                                'IsNewVersionOf',
                                'IsPreviousVersionOf',
                                'IsPartOf',
                                'HasPart',
                                'IsPublishedIn',
                                'IsReferencedBy',
                                'References',
                                'IsDocumentedBy',
                                'Documents',
                                'IsCompiledBy',
                                'Compiles',
                                'IsVariantFormOf',
                                'IsOriginalFormOf',
                                'IsIdenticalTo',
                                'IsReviewedBy',
                                'Reviews',
                                'IsDerivedFrom',
                                'IsSourceOf',
                                'IsRequiredBy',
                                'Requires',
                                'IsObsoletedBy',
                                'Obsoletes',
                            ],
                        },
                        'related_identifier_description': {'type': 'string'},
                        'resource_type_general': {'type': 'string'},  # TODO this isn't in the template to keep the template simple
                    },
                },
            },
            'examples': {
                'type': 'array',
                'minItems': 1,   # this pattern makes things self describing
                'items': {
                    'type': 'object',
                    'properties': {
                        'example_image_filename': {'type': 'string'},
                        'example_image_locator': {'type': 'string'},
                        'example_image_description': {'type': 'string'},
                    }
                }
            },
            'contributors': ContributorsExportSchema.schema,
        }
    }


class DatasetDescription2Schema(JSONSchema):
    # XXX not clear that we need a separate schema for this
    # or whether we handle it in the ingestion side, having
    # a single internal schema should be possible
    schema = {
        'errors': ErrorSchema.schema,
        'template_schema_version': {'type': 'string'},
        # 2.0
        # 'required': []
        'properties': {
            'dataset_type': {
                'type': 'string',
                'enum': ['experimental', 'computational', 'abi-scaffold', 'o2s-simulation'],
            },
            'title': {'type': 'string',},
            'subtitle': {'type': 'string',},
            'keywords': {'type': 'array',},
            'funding': {'type': 'array',},
            'acknowledgements': {'type': 'string',},

            'study': {
                'type': 'object',
                'properties': {
                    'study_purpose': {'type': 'string',},
                    'study_data_collection': {'type': 'string',},
                    'study_conclusion': {'type': 'string',},
                    'study_collection_title': {'type': 'string',},
                },},
            'study_organ_system': {'type': 'array',},
            'study_approach': {'type': 'array',},
            'study_technique': {'type': 'array',},

            'number_of_subjects': {'type': 'integer'},
            'number_of_samples': {'type': 'integer'},
        }
    }


DatasetDescriptionExportSchema.schema['properties'].update(
    DatasetDescription2Schema.schema['properties'])


class DatasetDescriptionSchema(JSONSchema):
    context = lambda : ({}, None)
    __schema = copy.deepcopy(DatasetDescriptionExportSchema.schema)
    schema = JApplyRecursive(
        EIS._to_pattern, __schema,
        alternates={'Orcid': idlib.Orcid._id_class.local_regex})


class SubmissionSchema(JSONSchema):
    context = lambda : ({}, None)
    schema = {
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'errors': ErrorSchema.schema,
            'submission': {
                'type': 'object',
                'required': ['sparc_award_number',
                             'milestone_achieved',
                             'milestone_completion_date'],
                # TODO allOf?
                'properties': {'sparc_award_number': {'type': 'string'},
                               'milestone_achieved': {'type': 'string'},
                               'milestone_completion_date': {'type': 'string',
                                                             'context_value': 'TEMP:milestoneCompletionDate',
                                                             'pattern': iso8601datepattern,},}}}}


class UnitSchema(JSONSchema):
    schema = {'oneOf': [{'type': ['number', 'string']},
                        {'type': 'object',
                         # TODO enum on the allowed values here
                         'properties': {'type': {'type': 'string'},
                                        'value': {'type': 'number'},
                                        'unit': {'type': 'string'}}}]}


_software_schema = {'type': 'array',
                    'minItems': 1,
                    'items': {
                        'type': 'object',
                        'properties': {
                            'software_version': {'type': 'string'},
                            'software_vendor': {'type': 'string'},
                            'software_url': {'type': 'string'},
                            'software_rrid': EIS._allOf(RridSchema),
                        },},}


subsam_common_properties = {  # FIXME these should not be in the samples sheet, or if they are they refer to the sample
    'species': {'anyOf': [EIS._allOf(OntTermSchema),
                          {'type': 'string'},],
                'context_value': 'sparc:animalSubjectIsOfSpecies'},
    'strain': {'anyOf': [EIS._allOf(OntTermSchema),
                         {'type': 'string'},],
                'context_value': 'sparc:animalSubjectIsOfStrain'},
    'rrid_for_strain': EIS._allOf(RridSchema, context_value=idtype('TEMP:specimenRRID')),  # XXX CHANGE from sparc:specimenHasIdentifier
    'genus': strcont('sparc:animalSubjectIsOfGenus'),
    'sex': {'anyOf': [EIS._allOf(OntTermSchema),
                      # FIXME oneOf fails when EIS is converted to input type
                      {'type': 'string'}],
            'context_value': idtype('sparc:hasBiologicalSex'),
            },  # sex as a variable ?

    'body_mass':{'context_value': 'sparc:animalSubjectHasWeight',
                 **UnitSchema.schema},
    'body_weight': {'context_value': 'sparc:animalSubjectHasWeight',
                    **UnitSchema.schema},
    'age': {'context_value': 'TEMP:hasAge',
            **UnitSchema.schema},
    'age_years': {'context_value': 'TEMP:hasAge',
                  **UnitSchema.schema},
    'age_category': {'type': 'string',
                     'context_value': idtype('TEMP:hasAgeCategory'),
                     },  # TODO uberon
    'age_range_min': {'context_value': 'TEMP:hasAgeMin',
                      **UnitSchema.schema},
    'age_range_max': {'context_value': 'TEMP:hasAgeMax',
                      **UnitSchema.schema},
    'handedness': strcont('TEMP:hasHandedness'),

    # ambiguous
    # genotype is ambiguous when dealing with tissue from chimeric subjects or experimentally manipulated cell populations
    'genotype': strcont('TEMP:hasGenotype'),
    'mass': {'context_value': 'sparc:animalSubjectHasWeight',
             # XXX NOTE subjects will require their own context
             # since mass can clearly have other meanings elsewhere
             **UnitSchema.schema},
    'weight': {'context_value': 'sparc:animalSubjectHasWeight',
               **UnitSchema.schema},
    'experimental_group': strcont('TEMP:hasAssignedGroup'),
    'group': strcont('TEMP:hasAssignedGroup'),
    'organism_rrid': strcont('TEMP:organismRRID'),

    # fields that require additional information
    'experiment_date': strcont('TEMP:protocolPerformanceDate'),  # FIXME needs to reference a protocol
    'injection_date': strcont('TEMP:protocolPerformanceDate'),  # FIXME needs to reference a protocol
    'date_of_euthanasia': strcont('TEMP:protocolPerformanceDate'),
    'date_of_injection': strcont('TEMP:protocolPerformanceDate'),

    # fields that are actually about files
    'sha1': strcont('TEMPRAW:hasDigitalArtifactThatIsAboutItWithHash'),
    'filename': strcont('TEMPRAW:hasDigitalArtifactThatIsAboutIt'),
    'upload_filename': strcont('TEMPRAW:hasDigitalArtifactThatIsAboutIt'),
    'dataset_id': strcont('TEMP:providerDatasetIdentifier'),  # FIXME need a global convention for this

    # fields that are actually about performances
    'experiment_number': strcont('TEMP:localPerformanceNumber'),  # FIXME TODO
    'session': strcont('TEMP:localPerformanceNumber'),

    # extra
    'comment': strcont('TEMP:providerNote'),
    'note': strcont('TEMP:providerNote'),
}


class PerformanceExportSchema(JSONSchema):
    schema = {
        'type': 'object',
        'jsonld_include': {'@type': ['sparc:Performance', 'owl:NamedIndividual']},
        '': [{'context_runtime': [
                '#/meta/uri_api',
                (lambda uri_api: uri_api + '/performances/'),
                '#/performances/@context/@base']},
             {'context_runtime': [
                 '#/meta/uri_api',
                 (lambda uri_api: {'@id': uri_api + '/performances/',
                                   '@prefix': True}),
                 '#/@context/performances']}],
        'required': ['performance_id',],
        'properties': {
            'performance_id': {'type': 'string',
                               'pattern': fs_safe_identifier_pattern,
                               'context_value': {"@id": "@id",
                                                 "@type": "@id"},
                           },
            'protocol_url_or_doi': _protocol_url_or_doi_schema,  # FIXME ideally this would be true but are we sure it always is?
        },}


class PerformancesExportSchema(JSONSchema):
    schema = {
        'type': 'object',
        #'context_value': {'@id': 'graph-subjects', CONTEXT_ROOT: None},  # not needed, schemas have their own context
        'required': ['performances'],
        'properties': {
            'performances': {
                'type': 'array',
                'minItems': 1,
                # FIXME there is currently no way to require that
                # a key be unique among all objects
                'items': PerformanceExportSchema.schema,},
            'errors': ErrorSchema.schema,}}


class PerformancesSchema(JSONSchema):
    context = lambda : ({}, None)
    __schema = copy.deepcopy(PerformancesExportSchema.schema)
    schema = JApplyRecursive(EIS._to_pattern, __schema)


class SubjectExportSchema(JSONSchema):
    schema = {
        'type': 'object',
        'jsonld_include': {'@type': ['sparc:Subject', 'owl:NamedIndividual']},
        '': [{'context_runtime': [
                '#/meta/uri_api',
                (lambda uri_api: uri_api + '/subjects/'),
                '#/subjects/@context/@base']},
             {'context_runtime': [
                 '#/meta/uri_api',
                 (lambda uri_api: {'@id': uri_api + '/subjects/',
                                   '@prefix': True}),
                 '#/@context/subjects']}],
        'required': ['subject_id', 'species'],
        'properties': {
            'subject_id': {'type': 'string',
                           'pattern': fs_safe_identifier_pattern,
                           'context_value': {"@id": "@id",
                                             "@type": "@id"},
                           #'context_runtime': [  # TODO -> derive after add ?
                               #'#/meta/uri_api',
                               # FIXME the @id mapping seems to break the @context !?
                                #(lambda uri_api: {
                                    #'@id': '@id',
                                    #'@context': {
                                        #'@base': uri_api + '/subjects/'}}),
                               #'#/subjects/@context/subject_id']
                           #'context_runtime':
                           #lambda base: {'@id': '@id',
                                         #'@type': '@id',
                                         #'@context': {'@base': f'{base}subjects/'}}
                           },
            'ear_tag_number': strcont('TEMP:localIdAlt'),  # FIXME also int? or int ids don't support mathematical operations
            'treatment': strcont('TEMP:hadExperimentalTreatmentApplied'),
            'initial_weight': cont('sparc:animalSubjectHasWeight', ['string', 'number']),
            'height_inches': cont('TEMP:subjectHasHeight', ['string', 'number']),  # FIXME units
            'gender': strcont('sparc:hasGender'),
            'body_mass_weight': cont('TEMP:subjectHasWeight', ['string', 'number']),  # FIXME units
            'anesthesia': strcont('TEMP:wasAdministeredAnesthesia'),
            'stimulation_site': strcont('sparc:spatialLocationOfModulator'),  # TODO ontology
            'stimulator': strcont('sparc:stimulatorUtilized'),
            'stimulating_electrode_type': strcont('sparc:stimulatorUtilized'),  # FIXME instance vs type

            'nerve': strcont('TEMPRAW:involvesAnatomicalRegion'),  # FIXME can be more specific than this
            **subsam_common_properties
        },}


class SubjectsExportSchema(JSONSchema):
    schema = {
        'type': 'object',
        #'context_value': {'@id': 'graph-subjects', CONTEXT_ROOT: None},  # not needed, schemas have their own context
        'required': ['subjects'],
        'properties': {
            'subjects': {
                'type': 'array',
                #'context_runtime': lambda base: {'@id': f'{base}subjects',},
                'minItems': 1,
                # FIXME there is currently no way to require that
                # a key be unique among all objects
                'items': SubjectExportSchema.schema,},
            'errors': ErrorSchema.schema,
            'software': _software_schema}}

    # FIXME not implemented
    extras = [['unique', ['subjects', '*', 'subject_id']]]


class SubjectsSchema(JSONSchema):
    context = lambda : ({}, None)
    __schema = copy.deepcopy(SubjectsExportSchema.schema)
    schema = JApplyRecursive(EIS._to_pattern, __schema)


class SampleExportSchema(JSONSchema):
    schema = {
        'type': 'object',
        'jsonld_include': {'@type': ['sparc:Sample', 'owl:NamedIndividual']},
        '': [{'context_runtime': [
                '#/meta/uri_api',
                (lambda uri_api: uri_api + '/samples/'),
                '#/samples/@context/@base']},
             {'context_runtime': [
                 '#/meta/uri_api',
                 (lambda uri_api: {'@id': uri_api + '/samples/',
                                   '@prefix': True}),
                 '#/@context/samples']}],
        'required': ['sample_id', 'subject_id'],
        'properties': {
            'sample_id': {'type': 'string',
                          'pattern': fs_safe_identifier_pattern,
                          },
            'subject_id': {'type': 'string',
                           'pattern': fs_safe_identifier_pattern,
                           },
            'primary_key': {'type': 'string',
                            'context_value': {"@id": "@id",
                                              "@type": "@id"},
                            #'context_runtime': [  # TODO -> derive after add ?
                                #'#/meta/uri_api',
                                #(lambda uri_api: uri_api + '/samples/'),
                                # FIXME the @id mapping seems to break the @context !?
                                # yep, this doesn't work because for some reason the
                                # @id has to use the parent space, but I think this
                                # approach to setting @base is probably safer
                                #(lambda uri_api: {
                                    #'@id': '@id',
                                    #'@context': {
                                        #'@base': uri_api + '/samples/'}}),
                                #'#/samples/@context/primary_key',
                                #'#/samples/@context/@base'],
                            #'context_runtime':
                            #lambda base: {'@id': '@id',
                                          #'@type': '@id',
                                          #'@context': {'@base': f'{base}samples/'}}
                            },
            'specimen': {'type': 'string',
                         },  # what are we expecting here?
            'specimen_anatomical_location': {'oneOf':
                                             [{'type': 'string',},
                                              # FIXME not seeing an easy wawy out of this
                                              # other than switching many of these over to arrays
                                              # by default
                                              {'type': 'array',
                                               'items': {'type': 'string'},},
                                              ],},

                        'disease': {'type': 'string',
                                    },
            'reference_atlas': {'type': 'string',
                                },
            'protocol_title': {'type': 'string',
                               },
            'protocol_url_or_doi': _protocol_url_or_doi_schema 
            ,
            'experimental_log_file_name': {'type': 'string',
                                           },
            **subsam_common_properties
        },}


class SamplesFileExportSchema(JSONSchema):
    schema = {
        'type': 'object',
        'required': ['samples'],
        'properties': {
            'samples': {
                'type': 'array',
                #'context_runtime': lambda base: {'@id': f'{base}samples',},
                'minItems': 1,
                'items': SampleExportSchema.schema,},
            'errors': ErrorSchema.schema,
            'software': _software_schema}}

    # FIXME not implemented
    extras = [['unique', ['samples', '*', 'sample_id']]]


class SamplesFileSchema(JSONSchema):
    context = lambda : ({}, None)
    __schema = copy.deepcopy(SamplesFileExportSchema.schema)
    schema = JApplyRecursive(EIS._to_pattern, __schema)


class ManifestRecordExportSchema(JSONSchema):
    schema = {
        'allOf': [
            {'type': 'object',
             'properties': {
                 'pattern': {
                     'type': 'string',
                 },
                 'filename': {
                     'type': 'string',
                 },
                 'timestamp': {
                     'type': 'string',
                     'pattern': iso8601pattern,
                 },
                 'timestamp_or_date': {  # this is internal for dataset template <= 1.2.3
                     'type': 'string',
                     'pattern': iso8601bothpattern,
                 },
                 'date': {
                     'type': 'string',
                     'pattern': iso8601datepattern,
                 },
                 'description': {
                     'type': 'string',
                 },
                 'file_type': {
                     'type': 'string',
                 },
                 'data_type': {
                     'type': 'string',
                 },
                 'additional_types': {
                     'type': 'string',  # FIXME currently expects a single value, maybe rename?
                     #'type': 'array',
                     #'minItems': 1,
                     #'items': {
                         #'type': 'string',
                         # FIXME need to make a note that if more than one
                         # value is provided then the value provided first
                         # is the one that will be lifted to mimetype for the file
                         # TODO mimetype pattern?
                     #},
                 },
                 # extras duplicated from other sheets
                 'protocl_title': {'type': 'string',},
                 'protocol_url_or_doi': _protocol_url_or_doi_schema,
                 'software_version': {'type': 'string'},
                 'software_vendor': {'type': 'string'},
                 'software_url': {'type': 'string'},
                 'software_rrid': EIS._allOf(RridSchema),
                 'software': {'type': 'string'},
                 # more extras
                 #'organism_RRID': EIS._allOf(RridSchema),
                 'sha1': {'type': 'string'},
             },
             },
            {'oneOf': [
                {'required': ['pattern']},
                {'required': ['filename']},  # timestamp not required for now
            ]},
        ]
    }


class ChecksumSchema(JSONSchema):
    _schema_base = {
        'type': 'object',
        'required': ['type', 'cypher'],
        'properties': {
            'type': {'type': 'string',
                     'enum': ('checksum',),},
            'cypher': {'type': 'string',
                       # more may be supported in the future
                       # but supported checksums must be closed
                       # because every single supported checksum
                       # requires a full implementation
                       'enum': ('sha256',),},
            # other options might be base64 or something like that
            'hex': {'type': 'string',},
        }
    }

    _schema_more = {'oneOf': [
        {'properties': {'cypher': {'type': 'string', 'enum': ('sha256',),},
                        'hex': {'type': 'string', 'minLength': 64, 'maxLength': 64,},},},
    ],}

    schema = {'allOf': [_schema_base,
                        _schema_more,]}


class PathSchema(JSONSchema):
    schema = {
        'type': 'object',
        'jsonld_include': {'@type': ['sparc:Path', 'owl:NamedIndividual']},
        'required': ['type',
                     'dataset_relative_path',
                     'uri_api',
                     'uri_human',
                     ],
        'properties': {
            'type': {'type': 'string',
                     'enum': ['path']},
            'dataset_relative_path': {'type': 'string'},
            'uri_api': {'oneOf': [{'type': 'string',
                                   'pattern': simple_url_pattern},
                                  {'type': 'string',
                                   'pattern': simple_file_url_pattern}]},
            'uri_human': {'oneOf': [{'type': 'string',
                                     'pattern': simple_url_pattern},
                                    {'type': 'string',
                                     'pattern': simple_file_url_pattern}]},
            'dataset_id': {'type': 'string'},
            'remote_id': {'type': 'string'},
            'parent_id': {'type': 'string'},
            'mimetype': {'type': 'string'},
            'magic_mimetype': {'type': 'string'},
            'contents': {'type': 'object'},  # opaque here
            'size_bytes': {'type': 'number'},
            'checksums': {'type': 'array',
                          'minItems': 1,
                          'items': ChecksumSchema.schema,},
            'basename': {'type': 'string',
                         # FIXME TODO this is overly restrictive and
                         # doesn't help in providing good feedback in
                         # cases where people are using periods as
                         # file name separators which messes up suffix
                         # file type detection, it also doesn't warn
                         # if there is no suffix
                         'pattern': fs_safe_identifier_pattern,},
            'errors': ErrorSchema.schema,
        }
    }


class PathTransitiveMetadataSchema(JSONSchema):
    schema = {'type': 'object',
              'properties': {
                  'type': {'type': 'string',
                           'enum': ('path-metadata',),},
                  'records': {'type': 'array',
                              'items': PathSchema.schema,
                              'minItems': 1,},},}


class ManifestFileExportSchema(JSONSchema):  # FIXME TODO FileObjectSchema ??
    schema = {
        'type': 'object',
        'jsonld_include': {'@type': ['sparc:File', 'owl:NamedIndividual']},
        # FIXME handle raw json manifest file structure where contents is the json blob
        # I think we just need to rename metadata -> manifest_records to fix that case
        'allOf': [
            PathSchema.schema,
            {'required': ['contents'],  # TODO uri_api
             'properties': {
                 'errors': ErrorSchema.schema,
                 'contents':  {
                     'type': 'object',
                     'required': ['manifest_records'],
                     'properties': {
                         'manifest_records': {
                             'type': 'array',
                             'minItems': 1,
                             'items': ManifestRecordExportSchema.schema,
                         },
                     },
                 },
             },
             },
        ]
    }


class ManifestFileSchema(JSONSchema):
    __schema = copy.deepcopy(ManifestFileExportSchema.schema)
    schema = JApplyRecursive(EIS._to_pattern, __schema)


class ManifestFilesExportSchema(JSONSchema):
    schema = {
        'type': 'object',
        'required': ['manifest_files'],
        'properties': {
            'manifest_files': {
                'type': 'array',
                'minItems': 1,
                'items': ManifestFileExportSchema.schema,
            },
            'errors': ErrorSchema.schema,
        }
    }


class MetaOutExportSchema(JSONSchema):
    __schema = copy.deepcopy(DatasetDescriptionExportSchema.schema)
    extra_required = ['award_number',
                      'principal_investigator',
                      'species',
                      'organ',
                      'modality',
                      'techniques',
                      'contributor_count',
                      'uri_human',  # from DatasetMetadata
                      'uri_api',  # from DatasetMetadata
                      'sparse',  # tells you whether files/dirs/size is reliable
                      'files',
                      'dirs',
                      'size',
                      'folder_name',  # from DatasetMetadat
                      'title',
                      'template_schema_version',
                      'number_of_subjects',
                      'number_of_samples',
                      'timestamp_created',
                      'timestamp_updated',
                      'timestamp_updated_contents',
                      #'subject_count',
                      #'sample_count',
    ]
    __schema['required'].remove('contributors')
    __schema['required'].remove('name')
    __schema['required'] += extra_required
    __schema['properties'].pop('contributors')
    __schema['properties'].pop('name')
    __schema['properties'].update({
        'errors': ErrorSchema.schema,
        'dirs': {'type': 'integer'},
        'doi': EIS._allOf(DoiSchema, context_value='TEMP:hasDoi'),
        'files': {'type': 'integer',
                  'context_value': inttype('TEMP:numberOfFiles'),
                  },
        'size': {'type': 'integer',
                 'context_value': inttype('TEMP:hasSizeInBytes'),
                 },
        'sparse': {'type': 'boolean',
                   'context_value': booltype('TEMP:sparseClone'),
                   },
        'folder_name': {'context_value': 'rdfs:label',
                        **string_noltws},
        'title': {'context_value': 'dc:title',
                  **string_noltws},
        'timestamp_created': {'type': 'string',
                              'pattern': iso8601pattern,
                              'context_value': 'TEMP:wasCreatedAtTime',
                              },
        'timestamp_updated': {'type': 'string',
                              'pattern': iso8601pattern,
                              'context_value': 'TEMP:wasUpdatedAtTime',
                              },
        'timestamp_updated_contents': {'type': 'string',
                                       'pattern': iso8601pattern,
                                       'context_value': 'TEMP:contentsWereUpdatedAtTime',
                                       },
        'uri_human': {'allOf': [  # FIXME this is absolutely not the right way to solve this problem
            {'type': 'string',
             'context_value': idtype('TEMP:hasUriHuman'),},
            {'oneOf': [
                # FIXME proper regex
                {'pattern': r'^https://app\.pennsieve\.io/N:organization:',},
                {'pattern': simple_file_url_pattern},]}]},
        'uri_api': {'allOf': [  # FIXME this is absolutely not the right way to solve this problem
            {'type': 'string',
             #'context_value': idtype('TEMP:hasUriApi'),
             'context_value': idtype('@id'),},
            {'oneOf': [
                # FIXME proper regex
                {'pattern': r'^https://api\.pennsieve\.io/(datasets|packages)/',},
                {'pattern': simple_file_url_pattern},]}]},
        'award_number': {'type': 'string',
                         'pattern': award_pattern,
                         'context_value': idtype('TEMP:hasAwardNumber', base=TEMP['awards/']),
                         },
        'principal_investigator': {'type': 'string',
                                   'context_value': idtype('TEMP:hasResponsiblePrincipalInvestigator'),
                                   },
        'protocol_url_or_doi': {'type': 'array',
                                'minItems': 1,
                                'items': _protocol_url_or_doi_schema,
                                'context_value': idtype('TEMP:hasProtocol'),
                                },
        'additional_links': {'type': 'array',
                             'minItems': 1,
                             'items': {
                                 'type': 'object',
                                 'properties': {
                                     'link': {'type': 'string'},
                                     'link_description': {'type': 'string'},
                                     },},
                             'context_value': idtype('TEMP:hasAdditionalLinks'),  # TODO
                             },
        'species': {'anyOf': [EIS._allOf(OntTermSchema),
                              {'type': 'string',},],
                    'context_value': idtype('isAbout'),  # FIXME vs has part generated from participant
                    },
        'template_schema_version': {'type': 'string'},
        'organ': {'type': 'array',
                  'minItems': 1,
                  'items': EIS._allOf(OntTermSchema),
                  'context_value': idtype('isAbout'),
                  },
        'modality': {'type': 'array',
                     'minItems': 1,
                     'items': {
                         'type': 'string',
                     },
                    'context_value': idtype('TEMP:hasExperimentalModality'),
        },
        'techniques': {'type': 'array',
                       'minItems': 1,
                       'items': EIS._allOf(OntTermSchema),
                       # FIXME usedProtocolThatEmployedTechniques ...
                       'context_value': idtype('TEMP:protocolEmploysTechnique'),
        },

        # TODO $ref these w/ pointer?
        # in contributor etc?

        'performance_count': {'type': 'integer',
                            'context_value': inttype('TEMP:hasNumberOfPerformances'),
                              },
        'subject_count': {'type': 'integer',
                          'context_value': inttype('TEMP:hasNumberOfSubjects'),
                          },
        'sample_count': {'type': 'integer',
                         'context_value': inttype('TEMP:hasNumberOfSamples'),
                         },
        'contributor_count': {'type': 'integer',
                              'context_value': inttype('TEMP:hasNumberOfSamples'),
                              },})

    schema = {
        'jsonld_include': {'@type': ['sparc:Dataset', 'owl:NamedIndividual']},
        'allOf': [__schema,
                  {'anyOf': [
                      #{'required': ['number_of_subjects']},
                      #{'required': ['number_of_samples']},
                            {'required': ['subject_count']},  # FIXME extract subjects from samples ?
                      {'required': ['sample_count']}
                  ]}]}


class MetaOutSchema(JSONSchema):
    context = lambda : ({}, None)
    __schema = copy.deepcopy(MetaOutExportSchema.schema)
    schema = JApplyRecursive(EIS._to_pattern, __schema)


class DatasetOutExportSchema(JSONSchema):
    """ Schema that adds the id field since investigators shouldn't need to know
        the id to check the integrity of their data. We need it because that is
        a key piece of information that we use to link everything together. """

    __schema = copy.deepcopy(DatasetStructureSchema.schema['allOf'][0])
    __schema['required'] = [
        'id',
        'meta',
        'contributors',
        'path_metadata',  # required for the time being as a sanity check on manifest structure, will move elsewhere in the future
        'prov']
    __schema['properties'] = {
        'id': {'type': 'string',  # ye old multiple meta/bf id issue
               'pattern': '^N:dataset:'},
        'meta': MetaOutExportSchema.schema,  # XXXXXXXXXXX wrong types ??
        #'readme': {'type': 'string',},  # TODO ...
        'rmeta': {'type': 'object',
                  # FIXME TODO not entirely sure how to deal with embedding remote metadata about
                  # datasets some of which also comes from this pipeline
                  # heh, recursive metadata ...
                  'properties': {'readme': {
                      'type': 'string',
                      'context_value': {'@id': 'TEMP:readmeText', '@type': 'ilxtr:Markdown'},
                  },},},
        'path_metadata': {'type': 'array'},
        'prov': ProvSchema.schema,
        'errors': ErrorSchema.schema,
        'contributors': ContributorsOutExportSchema.schema,
        'creators': CreatorsExportSchema.schema,
        'performances': PerformancesExportSchema.schema['properties']['performances'],  # 2.0.0
        # FIXME subjects_file might make the name overlap clearer
        'subjects': SubjectsExportSchema.schema['properties']['subjects'],  # FIXME SubjectsOutSchema
        'samples': SamplesFileExportSchema.schema['properties']['samples'],
        # 'identifier_metadata': {'type': 'array'},  # FIXME temporary
        'resources': {'type': 'array',
                      'items': {'type': 'object'},},
        'submission': {'type': 'object',},
        'specimen_dirs': {'type': 'object',  # FIXME unnest this
                          'required': ['records',],
                          'properties':
                          {'records': {'type': 'array',  # FIXME records vs contents we use both in manifests
                                       'items': {
                                           'type': 'object',
                                           'required': ['specimen_id', 'dirs'],
                                           'properties': {
                                               'type': {'type': 'string',  # FIXME seems wrong
                                                        'enum': ('SampleDirs', 'SubjectDirs'),},
                                               'specimen_id': {'type': 'string'},  # foreign key to subject or sample id
                                               'dirs': {'type': 'array',
                                                        'minItems': 1,},},},},
                           'errors': ErrorSchema.schema,},},
        'inputs': {'type': 'object',
                   # TODO do we need errors at this level?
                   'properties': {'dataset_description_file': DatasetDescriptionSchema.schema,
                                  'submission_file': SubmissionSchema.schema,
                                  'subjects_file': SubjectsSchema.schema,
                                  'samples_file':SamplesFileSchema.schema,
                                  },},}

    # FIXME switch to make samples optional since subject_id will always be there even in samples?
    schema = {
        'jsonld_include': {'@type': ['owl:Ontology']},
        '': [{'context_runtime': [  # induces indentation bug in Eamcs I think
                '#/meta/uri_api',
                (lambda uri_api: {'@id': uri_api + '/', '@prefix': True}),
                '#/@context/_pnc',]},
            {'context_runtime': [
                '#/meta/uri_api',
                (lambda uri_api: uri_api + '/'),
                '#/@context/@base',]}],
        'allOf': [__schema,
                  {'anyOf': [
                      {'required': ['subjects']},  # FIXME extract subjects from samples ?
                      {'required': ['samples']},
                      {'properties': {
                          'meta': {'properties': {
                              'number_of_subjects': {'type': 'integer',
                                                     'minimum': 0,
                                                     'maximum': 0,
                                                     },
                              'number_of_samples': {'type': 'integer',
                                                    'minimum': 0,
                                                    'maximum': 0,
                                                    },
                          }}}},
                  ]}]}


class DatasetOutSchema(JSONSchema):
    context = lambda : ({}, None)
    __schema = copy.deepcopy(DatasetOutExportSchema.schema)
    schema = JApplyRecursive(EIS._to_pattern, __schema)


class StatusSchema(JSONSchema):
    schema = {'type': 'object',
              'required': ['submission_index', 'curation_index', 'error_index',
                           'submission_errors', 'curation_errors'],
              'properties': {
                  'status_on_platform': {'type': 'object',  # NOTE becomes a json literal iirc
                                         'context_value': 'TEMP:remotePlatformStatus',
                                         },
                  'submission_index': {'type': 'integer', 'minimum': 0,
                                       'context_value': inttype('TEMP:submissionIndex'),
                                       },
                  'curation_index': {'type': 'integer', 'minimum': 0,
                                     'context_value': inttype('TEMP:curationIndex'),
                                     },
                  'error_index': {'type': 'integer', 'minimum': 0,
                                  'context_value': inttype('TEMP:errorIndex'),
                                  },
                  'submission_errors':{'type':'array',  # allow these to be empty
                                       'items': {'type': 'object'},
                                       },
                  'curation_errors':{'type':'array',  # allow these to be empty
                                     'items': {'type': 'object'},
                                     },
              }}


class PostSchema(JSONSchema):
    """ This is used to validate only the status schema part that is added to the DatasetOutSchema """
    schema = {
        'type': 'object',
        'required': [
            #'path_metadata',  # issue resolved for now determined to be non-fatal
            'status',
        ],
        'properties': {
            'path_metadata': {'type': 'array'},
            'status': StatusSchema.schema,},
    }


class SummarySchema(JSONSchema):
    # TODO add expected counts
    schema = {'type': 'object',
              'required': ['id', 'meta', 'prov'],
              'properties': {'id': {'type': 'string',
                                    'oneOf': [
                                        {'pattern': '^.*/.*$'},  # a file path based id
                                        {'pattern': '^N:organization:'},
                                    ],},
                             'meta': {'type': 'object',
                                      # folder name cannot be required when combining from existing exports
                                      'required': ['count', 'uri_api', 'uri_human'],
                                      'properties': {'folder_name': {'type': 'string'},
                                                     # FIXME common source for these with MetaOutSchema
                                                     'uri_human': {'type': 'string',
                                                                   'oneOf': [
                                                                       {'pattern': simple_file_url_pattern},
                                                                       {'pattern': r'^https://app\.pennsieve\.io/N:organization:'},
                                                                   ],},
                                                     'uri_api': {'type': 'string',
                                                                 'oneOf': [
                                                                     {'pattern': simple_file_url_pattern},
                                                                     {'pattern': r'^https://api\.pennsieve\.io/organizations/N:organization:'},
                                                                 ],},
                                                     'count': {'type': 'integer'},},},
                             'prov': {'type': 'object'},
                             'datasets': {'type': 'array',
                                          'minItems': 1,
                                          'items': {'type': 'object'},  # DatasetOutSchema already validated
                             }}}


class HeaderSchema(JSONSchema):
    schema = {'type': 'array',
              'items': {'type': 'string'}}


# constraints on MIS classes
class MISSpecimenSchema(JSONSchema):
    schema = {
        'type': 'object',
        'properties': {
            'TEMP:hasAge': {},
            'TEMP:hasAgeCategory': {},  # hasDevelopmentalStage
            'TEMP:hasBiologicalSex': {},  # TEMP because sparc:hasPhenotypicSex
            'TEMP:hasWeight': {},
            'TEMP:hasGenus': {},
            'TEMP:hasSpecies': {},
            'TEMP:hasStrain': {},  # race/ethnicity?
            'TEMP:hasRRID': {},
            'TEMP:hasGeneticVariation': {},

            'sparc:hasBodyMassIndex': {} ,
            'sparc:hasClinicalInformation': {},
            'sparc:hasDiseaseInformation': {},
            'sparc:hasDrugInformation': {},
            'sparc:hasFamilyHistory': {},
            'sparc:hasGender': {},
            'TEMP:hasHandedness': {},
            'TEMP:hasRace': {},
            'TEMP:hasEthnicity': {},
        },
        'oneOf': [
            {'allOf': [
                {'allOf':
                 [{'required': [
                     'TEMP:hasSpecies',
                     'TEMP:hasBiologicalSex',
                     'TEMP:hasWeight',]},
                  {'oneOf': [{'required': ['TEMP:hasAge']},
                             {'required': ['TEMP:hasAgeCategory']}]}]},
                {'oneOf': [
                    {'properties': {'TEMP:hasSpecies': {'enum': ['NCBITaxon:9606']}},
                     'required': [
                         'sparc:hasBodyMassIndex',
                         'sparc:hasClinicalInformation',
                         'sparc:hasDiseaseInformation',
                         'sparc:hasDrugInformation',
                         'sparc:hasFamilyHistory',
                         #'sparc:hasGender',
                         #'TEMP:hasHandedness',
                         'TEMP:hasRace',
                         'TEMP:hasEthnicity',
                     ]},
                    {'allOf': [
                        {'not': {'properties': {'TEMP:hasSpecies': {'enum': ['NCBITaxon:9606']}}}},
                        {'required': [
                            #'TEMP:hasSupplier',
                            #'TEMP:hasStrain',  # RRID? race/ethnicity?
                            'TEMP:hasRRID',
                         ]},
                    ]},
                ]},
            ]},
        ],
    }


class MISDatasetSchema(JSONSchema):
    moves = [  # FIXME this goes somehwere else
        # from datasetoutschema -> mis
        [[''], ['']],
        [[''], ['']],
        [[''], ['']],
        [[''], ['']],
        [[''], ['']],
        [[''], ['']],
        [['resources'], ['']],
        [['meta', 'protocol_url_or_doi'], ['']],
        #[['meta'], ['sparc:Resource']],
        [['subjects'], ['sparc:Specimen']],
        [['contributors'], ['sparc:Researcher']],  # ah the role confusion
    ]
    schema = {'type': 'object',
              'required': ['curies', 'rdf:type'],
              'properties': {
                  'curies': {'type': 'object'},
                  'rdf:type': {'type': 'string',
                               'pattern': 'sparc:Resource'},
                  #'sparc:'
                  'sparc:Specimen': {'type': 'array',
                                     'items': MISSpecimenSchema.schema},
              }
              }


runtime_context_specs = get_runtime_context_specs()
